# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Aniek Roelofs

import json
import logging
from datetime import datetime
from typing import Union

from airflow.exceptions import AirflowException
from airflow.hooks.base import BaseHook
from airflow.models.dag import DAG
from airflow.models.dagbag import DagBag
from airflow.models.dagrun import DagRun
from airflow.models.taskinstance import TaskInstance
from airflow.models.variable import Variable
from croniter import croniter

from observatory.platform.observatory_config import TerraformConfig, VirtualMachine
from observatory.platform.terraform_api import TerraformApi
from observatory.platform.utils.airflow_utils import (
    AirflowConns,
    AirflowVars,
    change_task_log_level,
    check_connections,
    check_variables,
    create_slack_webhook,
)


def get_workspace_id() -> str:
    """Uses terraform API and workspace name to get the id of this workspace.

    :return: workspace id
    """
    token = BaseHook.get_connection(AirflowConns.TERRAFORM).password
    terraform_api = TerraformApi(token)

    # Get organization
    organization = Variable.get(AirflowVars.TERRAFORM_ORGANIZATION)

    # Get workspace
    environment = Variable.get(AirflowVars.ENVIRONMENT)
    workspace = TerraformConfig.WORKSPACE_PREFIX + environment

    # Get workspace ID
    workspace_id = terraform_api.workspace_id(organization, workspace)

    return workspace_id


def get_last_execution_prev(dag: DAG, dag_id: str, prev_start_time_vm: Union[datetime, None]) -> Union[datetime, None]:
    """Find the execution date of the last DAG run before the previous time the VM was turned on.
    If there aren't any DAG runs before this time or the time is None (first/second time turning off VM) the
    execution date is set to the start_date of the DAG instead.

    If a DAG is currently running it will return None and the remaining tasks are skipped.

    :param dag: DAG object
    :param dag_id: the dag id
    :param prev_start_time_vm: previous time the VM was turned on
    :return: execution date or None
    """

    # Get execution date of the last run before previous start date
    all_dag_runs = DagRun.find(dag_id=dag_id)
    # sort dag runs by start datetime, newest first
    for dag_run in sorted(all_dag_runs, key=lambda x: x.start_date, reverse=True):
        if dag_run.state == "running":
            logging.info("DAG is currently running.")
            return None
        # None if first time running destroy
        if prev_start_time_vm:
            if dag_run.start_date < prev_start_time_vm:
                # get execution date of last run from when the VM was previously on
                return dag_run.execution_date
    else:
        # No runs executed previously
        if prev_start_time_vm:
            logging.info("No DAG runs that started before prev_start_time_vm.")
        else:
            # First time running destroy_vm, no previous start date available
            logging.info("No prev_start_time_vm.")
        logging.info("Setting last execution date to start_date of DAG.")
        last_execution_prev = dag.default_args["start_date"]
    return last_execution_prev


def check_success_runs(dag_id: str, execution_dates: list) -> bool:
    """For each date in the execution dates it checks if a DAG run exists and if so if the state is set to success.

    Only if both of these are true for all dates it will return True.

    :param dag_id: the dag id
    :param execution_dates: list of execution dates
    :return: True or False
    """
    for date in execution_dates:
        dag_runs = DagRun.find(dag_id=dag_id, execution_date=date)
        if not dag_runs:
            logging.info(f"Expected dag run on {date} has not been scheduled yet")
            return False

        for dag_run in dag_runs:
            logging.info(
                f"id: {dag_run.dag_id}, start date: {dag_run.start_date}, execution date: "
                f"{dag_run.execution_date}, state: {dag_run.state}"
            )
            if dag_run.state != "success":
                return False
    return True


class TerraformTasks:
    """A container for holding the functions for the Terraform telescope. Both DAGs vm_create and vm_destroy use
    functions from this telescope."""

    # DAG variables
    DAG_ID_CREATE_VM = "vm_create"
    DAG_ID_DESTROY_VM = "vm_destroy"

    XCOM_START_TIME_VM = "start_time_vm"
    XCOM_PREV_START_TIME_VM = "prev_start_time_vm"
    XCOM_DESTROY_TIME_VM = "destroy_time_vm"
    XCOM_TERRAFORM_RUN_ID = "terraform_run_id"
    XCOM_WARNING_TIME = "last_warning_time"

    TASK_ID_CHECK_DEPENDENCIES = "check_dependencies"
    TASK_ID_VM_STATUS = "check_vm_status"
    TASK_ID_VAR_UPDATE = "update_terraform_var"
    TASK_ID_RUN = "terraform_run_configuration"
    TASK_ID_DAG_SUCCESS = "check_all_dags_success"
    TASK_ID_RUN_STATUS = "terraform_check_run_status"
    TASK_ID_VM_RUNTIME = "check_runtime_vm"

    # Name of variable in terraform configuration that will be updated
    TERRAFORM_CREATE_VM_KEY = "airflow_worker_vm"
    TERRAFORM_CREATE_KEY = "create"

    # Name of module in terraform configuration that will be targeted
    TERRAFORM_MODULE_WORKER_VM = "module.airflow_worker_vm"

    # No. hours for VM to be turned on until a slack notification is send
    VM_RUNTIME_H_WARNING = 20

    # Frequency in hours of how often a warning message is sent on slack
    WARNING_FREQUENCY_H = 5

    @staticmethod
    def check_dependencies():
        """Check that all variables exist that are required to run the DAG.
        :return: None.
        """

        vars_valid = check_variables(
            AirflowVars.PROJECT_ID, AirflowVars.TERRAFORM_ORGANIZATION, AirflowVars.VM_DAGS_WATCH_LIST
        )
        conns_valid = check_connections(AirflowConns.TERRAFORM)

        if not vars_valid or not conns_valid:
            raise AirflowException("Required variables or connections are missing")

    @staticmethod
    def get_variable_create(**kwargs) -> bool:
        """Retrieves the current value of the terraform variable from the cloud workspace, if this already has the
        value
        that the DAG is planning to set it to, it will return False and remaining tasks are skipped.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed to
        this argument.
        :return: True or False
        """

        token = BaseHook.get_connection(AirflowConns.TERRAFORM).password
        terraform_api = TerraformApi(token)

        # Add variable to workspace
        workspace_id = get_workspace_id()
        variables = terraform_api.list_workspace_variables(workspace_id)
        vm = None
        for var in variables:
            if var.key == TerraformTasks.TERRAFORM_CREATE_VM_KEY:
                vm = VirtualMachine.from_hcl(var.value)
                break

        # create_vm dag run and terraform create variable is already true (meaning VM is already on) or
        # destroy_vm dag run and terraform destroy variable is already false (meaning VM is already off)
        vm_is_on = vm.create
        print(f"DAG RUN: {kwargs['dag_run'].dag_id}")
        print(f"VM state: {vm_is_on}")

        if (kwargs["dag_run"].dag_id == TerraformTasks.DAG_ID_CREATE_VM and vm_is_on) or (
            kwargs["dag_run"].dag_id == TerraformTasks.DAG_ID_DESTROY_VM and not vm_is_on
        ):
            logging.info(f'VM is already in this state: {"on" if vm_is_on else "off"}')
            return False

        logging.info(f'Turning vm {"off" if vm.create else "on"}')
        return True

    @staticmethod
    def update_terraform_variable(**kwargs):
        """Updates the value of the terraform variable from the cloud workspace. The value is dependent on which DAG
        this function is executed from.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed to
        this argument.
        :return: None
        """

        create_value = kwargs["dag_run"].dag_id == TerraformTasks.DAG_ID_CREATE_VM
        token = BaseHook.get_connection(AirflowConns.TERRAFORM).password
        terraform_api = TerraformApi(token)

        # Get variable
        workspace_id = get_workspace_id()
        variables = terraform_api.list_workspace_variables(workspace_id)
        vm = None
        vm_var = None
        for var in variables:
            if var.key == TerraformTasks.TERRAFORM_CREATE_VM_KEY:
                vm_var = var
                vm = VirtualMachine.from_hcl(var.value)
                break

        # Update vm create value and convert to HCL
        vm.create = create_value
        print(f"CREATE VALUE!: {vm.create}")
        vm_var.value = vm.to_hcl()

        # Update value
        terraform_api.update_workspace_variable(vm_var, workspace_id)

    @staticmethod
    def terraform_run(**kwargs):
        """Runs terraform configuration on specific target.
        xcom values are pushed for the start and destroy time of the vm with the start time of this task. If this
        function is executed from the 'create' DAG, the time from when the previous vm was started will be set to the
        xcom value of this task from the previous run.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed to
        this argument.
        :return: None
        """

        # Push xcom with start date of this DAG run both for start and destroy
        ti: TaskInstance = kwargs["ti"]
        if kwargs["dag_run"].dag_id == TerraformTasks.DAG_ID_CREATE_VM:
            prev_start_time_vm = ti.xcom_pull(
                key=TerraformTasks.XCOM_START_TIME_VM,
                task_ids=TerraformTasks.TASK_ID_RUN,
                dag_id=TerraformTasks.DAG_ID_CREATE_VM,
                include_prior_dates=True,
            )
            ti.xcom_push(TerraformTasks.XCOM_PREV_START_TIME_VM, prev_start_time_vm)
            ti.xcom_push(TerraformTasks.XCOM_START_TIME_VM, ti.start_date)
        if kwargs["dag_run"].dag_id == TerraformTasks.DAG_ID_DESTROY_VM:
            ti.xcom_push(TerraformTasks.XCOM_DESTROY_TIME_VM, ti.start_date)

        token = BaseHook.get_connection(AirflowConns.TERRAFORM).password
        terraform_api = TerraformApi(token)

        target_addrs = TerraformTasks.TERRAFORM_MODULE_WORKER_VM
        workspace_id = get_workspace_id()
        message = f'Triggered from airflow DAG "{kwargs["dag_run"].dag_id}" at {ti.start_date}'

        run_id = terraform_api.create_run(workspace_id, target_addrs, message)
        logging.info(run_id)

        # Push run id
        ti.xcom_push(TerraformTasks.XCOM_TERRAFORM_RUN_ID, run_id)

    @staticmethod
    def check_terraform_run_status(**kwargs):
        """Retrieve the terraform run status until it is in a finished state, either successful or errored. See
        https://www.terraform.io/docs/cloud/api/run.html for possible run_status values.
        If the run status is not successful and the environment isn't develop a warning message will be sent to a slack
        channel.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed to
        this argument.
        :return: None
        """

        ti: TaskInstance = kwargs["ti"]
        run_id = ti.xcom_pull(key=TerraformTasks.XCOM_TERRAFORM_RUN_ID, task_ids=TerraformTasks.TASK_ID_RUN)
        project_id = Variable.get(AirflowVars.PROJECT_ID)

        token = BaseHook.get_connection(AirflowConns.TERRAFORM).password
        terraform_api = TerraformApi(token)

        run_status = None
        while run_status not in [
            "planned_and_finished",
            "applied",
            "errored",
            "discarded",
            "canceled",
            "force_canceled",
        ]:
            run_details = terraform_api.get_run_details(run_id)
            run_status = run_details["data"]["attributes"]["status"]

        logging.info(f"Run status: {run_status}")
        comments = f"Terraform run status: {run_status}"
        logging.info(f'Sending slack notification: "{comments}"')
        slack_hook = create_slack_webhook(comments, project_id, **kwargs)
        slack_hook.execute()

    @staticmethod
    def check_success_dags(**kwargs) -> str:
        """Check if all expected runs for the DAGs in the watchlist are successful. If they are the task to update the
        terraform variable will be executed, otherwise the task to check how long the VM has been on will be executed.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed to
        this argument.
        :return: id of task which should be executed next
        """

        destroy_worker_vm = True

        ti: TaskInstance = kwargs["ti"]
        prev_start_time_vm = ti.xcom_pull(
            key=TerraformTasks.XCOM_PREV_START_TIME_VM,
            task_ids=TerraformTasks.TASK_ID_RUN,
            dag_id=TerraformTasks.DAG_ID_CREATE_VM,
            include_prior_dates=True,
        )
        start_time_vm = ti.xcom_pull(
            key=TerraformTasks.XCOM_START_TIME_VM,
            task_ids=TerraformTasks.TASK_ID_RUN,
            dag_id=TerraformTasks.DAG_ID_CREATE_VM,
            include_prior_dates=True,
        )
        destroy_time_vm = ti.xcom_pull(
            key=TerraformTasks.XCOM_DESTROY_TIME_VM,
            task_ids=TerraformTasks.TASK_ID_RUN,
            dag_id=TerraformTasks.DAG_ID_DESTROY_VM,
            include_prior_dates=True,
        )
        logging.info(
            f"prev_start_time_vm: {prev_start_time_vm}, start_time_vm: {start_time_vm}, "
            f"destroy_time_vm: {destroy_time_vm}\n"
        )

        # Load VM DAGs watch list
        vm_dags_watch_list_str = Variable.get(AirflowVars.VM_DAGS_WATCH_LIST)
        logging.info(f"vm_dags_watch_list_str str: {vm_dags_watch_list_str}")
        vm_dags_watch_list = json.loads(vm_dags_watch_list_str)
        logging.info(f"vm_dags_watch_list: {vm_dags_watch_list}")

        for dag_id in vm_dags_watch_list:
            dagbag = DagBag()
            dag = dagbag.get_dag(dag_id)
            logging.info(f"Dag id: {dag_id}")

            # vm turned on manually and never turned on before
            if not start_time_vm and not prev_start_time_vm:
                logging.warning("Both start_time_vm and prev_start_time_vm are None. Unsure whether to turn off DAG.")
                destroy_worker_vm = False
                break

            # returns last execution date of previous vm cycle or None if a DAG is running
            last_execution_prev = get_last_execution_prev(dag, dag_id, prev_start_time_vm)
            if not last_execution_prev:
                destroy_worker_vm = False
                break

            logging.info(f"Execution date of last DAG before prev_start_time_vm: {last_execution_prev}\n")

            if destroy_time_vm:
                if start_time_vm < destroy_time_vm:
                    # If the vm is on, but there's no start_time_vm it must have been turned on manually.
                    logging.warning(
                        "start_time_vm is before destroy_time_vm. Perhaps the vm was turned on "
                        "manually in between. This task will continue with the given start_time_vm."
                    )

            # get a backfill of all expected runs between last execution date of prev cycle and the time the vm was
            # started. create iterator starting at the latest planned schedule before start_time_vm
            cron_iter = croniter(dag.normalized_schedule_interval, dag.previous_schedule(start_time_vm))

            # the last_execution_current is expected 1 schedule interval before the DAGs 'previous_schedule',
            # because airflow won't trigger a DAG until 1 schedule interval after the 'execution_date'.
            last_execution_current = cron_iter.get_prev(datetime)

            # if DAG is not set to catchup any backfill, the only run date is the last one.
            if dag.catchup:
                execution_dates = dag.get_run_dates(last_execution_prev, last_execution_current)
            else:
                execution_dates = [last_execution_current]

            # for each execution date check if state is success. This can't be done in all_dag_runs above, because the
            # dag_run might not be in all_dag_runs yet, because it is not scheduled yet.
            destroy_worker_vm = check_success_runs(dag_id, execution_dates)
            if destroy_worker_vm is False:
                break

        logging.info(f"Destroying worker VM: {destroy_worker_vm}")
        if destroy_worker_vm:
            return TerraformTasks.TASK_ID_VAR_UPDATE
        else:
            return TerraformTasks.TASK_ID_VM_RUNTIME

    @staticmethod
    def check_runtime_vm(**kwargs):
        """Checks how long the VM has been turned on based on the xcom value from the terraform run task.
        A warning message will be send in a slack channel if it has been on longer than the warning limit,
        the environment isn't develop and a message hasn't been send already in the last x hours.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are passed to
        this argument.
        :return: None
        """

        ti: TaskInstance = kwargs["ti"]
        last_warning_time = ti.xcom_pull(
            key=TerraformTasks.XCOM_WARNING_TIME,
            task_ids=TerraformTasks.TASK_ID_VM_RUNTIME,
            dag_id=TerraformTasks.DAG_ID_DESTROY_VM,
            include_prior_dates=True,
        )
        start_time_vm = ti.xcom_pull(
            key=TerraformTasks.XCOM_START_TIME_VM,
            task_ids=TerraformTasks.TASK_ID_RUN,
            dag_id=TerraformTasks.DAG_ID_CREATE_VM,
            include_prior_dates=True,
        )

        if start_time_vm:
            # calculate number of hours passed since start time vm and now
            hours_on = (ti.start_date - start_time_vm).total_seconds() / 3600
            logging.info(
                f"Start time VM: {start_time_vm}, hours passed since start time: {hours_on}, warning limit: "
                f"{TerraformTasks.VM_RUNTIME_H_WARNING}"
            )

            # check if a warning has been sent previously and if so, how many hours ago
            if last_warning_time:
                hours_since_warning = (ti.start_date - last_warning_time).total_seconds() / 3600
            else:
                hours_since_warning = None

            # check if the VM has been on longer than the limit
            if hours_on > TerraformTasks.VM_RUNTIME_H_WARNING:
                #  check if no warning was sent before or last time was longer ago than warning frequency
                if not hours_since_warning or hours_since_warning > TerraformTasks.WARNING_FREQUENCY_H:
                    comments = (
                        f"Worker VM has been on since {start_time_vm}. No. hours passed since then: "
                        f"{hours_on}."
                        f" Warning limit: {TerraformTasks.VM_RUNTIME_H_WARNING}H"
                    )
                    project_id = Variable.get(AirflowVars.PROJECT_ID)
                    slack_hook = create_slack_webhook(comments, project_id, **kwargs)

                    # http_hook outputs the secret token, suppressing logging 'info' by setting level to 'warning'
                    old_levels = change_task_log_level(logging.WARNING)
                    slack_hook.execute()
                    # change back to previous levels
                    change_task_log_level(old_levels)

                    ti.xcom_push(TerraformTasks.XCOM_WARNING_TIME, ti.start_date)
        else:
            logging.info(f"Start time VM unknown.")



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Aniek Roelofs, James Diprose, Tuan Chien

import logging
from datetime import timedelta
from typing import Dict

import pendulum
from airflow.exceptions import AirflowSkipException
from airflow.models.taskinstance import TaskInstance
from google.cloud.bigquery import SourceFormat
from observatory.platform.utils.airflow_utils import AirflowVars
from observatory.platform.utils.workflow_utils import (
    batch_blob_name,
    blob_name,
    bq_append_from_file,
    bq_append_from_partition,
    bq_delete_old,
    bq_load_ingestion_partition,
    table_ids_from_path,
    upload_files_from_list,
)
from observatory.platform.workflows.workflow import Release, Workflow


class StreamRelease(Release):
    def __init__(
        self,
        dag_id: str,
        start_date: pendulum.DateTime,
        end_date: pendulum.DateTime,
        first_release: bool,
        download_files_regex: str = None,
        extract_files_regex: str = None,
        transform_files_regex: str = None,
    ):
        """Construct a StreamRelease instance
        :param dag_id: the id of the DAG.
        :param start_date: the start_date of the release.
        :param end_date: the end_date of the release.
        :param first_release: whether this is the first release that is processed for this DAG
        :param download_files_regex: regex pattern that is used to find files in download folder
        :param extract_files_regex: regex pattern that is used to find files in extract folder
        :param transform_files_regex: regex pattern that is used to find files in transform folder
        """
        self.start_date = start_date
        self.end_date = end_date
        self.first_release = first_release
        release_id = self.start_date.strftime("%Y_%m_%d") + "-" + self.end_date.strftime("%Y_%m_%d")
        super().__init__(dag_id, release_id, download_files_regex, extract_files_regex, transform_files_regex)


class StreamTelescope(Workflow):
    def __init__(
        self,
        dag_id: str,
        start_date: pendulum.DateTime,
        schedule_interval: str,
        dataset_id: str,
        merge_partition_field: str,
        bq_merge_days: int,
        schema_folder: str,
        catchup: bool = False,
        queue: str = "default",
        max_retries: int = 3,
        max_active_runs: int = 1,
        source_format: SourceFormat = SourceFormat.NEWLINE_DELIMITED_JSON,
        schema_prefix: str = "",
        schema_version: str = None,
        load_bigquery_table_kwargs: Dict = None,
        dataset_description: str = "",
        table_descriptions: Dict[str, str] = None,
        batch_load: bool = False,
        airflow_vars: list = None,
        airflow_conns: list = None,
    ):
        """Construct a StreamTelescope instance.

        :param dag_id: the id of the DAG.
        :param start_date: the start date of the DAG.
        :param schedule_interval: the schedule interval of the DAG.
        :param dataset_id: the dataset id.
        :param merge_partition_field: the BigQuery field used to match partitions for a merge
        :param bq_merge_days: how often partitions should be merged (every x days)
        :param schema_folder: the path to the SQL schema folder.
        :param catchup: whether to catchup the DAG or not.
        :param queue: the Airflow queue name.
        :param max_retries: the number of times to retry each task.
        :param max_active_runs: the maximum number of DAG runs that can be run at once.
        :param source_format: the format of the data to load into BigQuery.
        :param schema_prefix: the prefix used to find the schema path
        :param schema_version: the version used to find the schema path
        :param load_bigquery_table_kwargs: the customisation parameters for loading data into a BigQuery table.
        :param dataset_description: description for the BigQuery dataset.
        :param table_descriptions: a dictionary with table ids and corresponding table descriptions
        :param batch_load: whether all files in the transform folder are loaded into 1 table at once
        :param airflow_vars: list of airflow variable keys, for each variable it is checked if it exists in airflow
        :param airflow_conns: list of airflow connection keys, for each connection it is checked if it exists in airflow
        """

        # Set transform_bucket_name as required airflow variable
        if not airflow_vars:
            airflow_vars = []
        airflow_vars = list(set([AirflowVars.TRANSFORM_BUCKET] + airflow_vars))

        super().__init__(
            dag_id,
            start_date,
            schedule_interval,
            catchup=catchup,
            queue=queue,
            max_retries=max_retries,
            max_active_runs=max_active_runs,
            airflow_vars=airflow_vars,
            airflow_conns=airflow_conns,
        )

        self.schema_prefix = schema_prefix
        self.schema_version = schema_version
        self.dataset_id = dataset_id
        self.source_format = source_format
        self.merge_partition_field = merge_partition_field
        self.bq_merge_days = bq_merge_days
        self.schema_folder = schema_folder
        self.load_bigquery_table_kwargs = load_bigquery_table_kwargs if load_bigquery_table_kwargs else dict()
        self.dataset_description = dataset_description
        self.table_descriptions = table_descriptions if table_descriptions else dict()
        self.batch_load = batch_load

    def get_release_info(self, **kwargs) -> bool:
        """Create a release instance and update the xcom value with the last start date.
        :param kwargs: The context passed from the PythonOperator.
        :return: None.
        """
        ti: TaskInstance = kwargs["ti"]

        first_release = False
        release_info = ti.xcom_pull(key=self.RELEASE_INFO, include_prior_dates=True)
        if not release_info:
            first_release = True
            # set start date to the start of the DAG
            start_date = pendulum.instance(kwargs["dag"].default_args["start_date"]).start_of("day")
        else:
            # set start date to end date of previous DAG run, add 1 day, because end date was processed in prev run.
            start_date = pendulum.parse(release_info[1]) + timedelta(days=1)
        # set start date to current day, subtract 1 day, because data from same day might not be available yet.
        end_date = pendulum.today("UTC") - timedelta(days=1)
        logging.info(f"Start date: {start_date}, end date: {end_date}, first release: {first_release}")

        # Turn dates into strings.  Prefer JSON'able data over pickling in Airflow 2.
        start_date = start_date.format("YYYYMMDD")
        end_date = end_date.format("YYYYMMDD")

        ti.xcom_push(self.RELEASE_INFO, (start_date, end_date, first_release))
        return True

    def download(self, release: StreamRelease, **kwargs):
        """Task to download the StreamRelease release.

        :param release: a StreamRelease instance.
        :param kwargs: The context passed from the PythonOperator.
        :return: None.
        """
        release.download()

    def upload_downloaded(self, release: StreamRelease, **kwargs):
        """Upload the downloaded files for the given release.

        :param release: a StreamRelease instance
        :param kwargs: The context passed from the PythonOperator.
        :return: None.
        """
        upload_files_from_list(release.download_files, release.download_bucket)

    def extract(self, release: StreamRelease, **kwargs):
        """Task to extract the StreamRelease release.

        :param release: a StreamRelease instance.
        :param kwargs: The context passed from the PythonOperator.
        :return: None.
        """
        release.extract()

    def transform(self, release: StreamRelease, **kwargs):
        """Task to transform the StreamRelease release.

        :param release: a StreamRelease instance.
        :param kwargs: The context passed from the PythonOperator.
        :return: None.
        """
        release.transform()

    def upload_transformed(self, release: StreamRelease, **kwargs):
        """Upload the transformed files for the given release.
        :param release: a StreamRelease instance
        :param kwargs: The context passed from the PythonOperator.
        :return: None.
        """
        upload_files_from_list(release.transform_files, release.transform_bucket)

    def bq_load_partition(self, release: StreamRelease, **kwargs):
        """For each file listed in transform_files, create a time based partition in BigQuery
        :param release: a StreamRelease instance
        :param kwargs: The context passed from the PythonOperator.
        :return: None.
        """
        if release.first_release:
            # The main table is the same as the partition, so no need to upload the partition as well. Especially
            # because a first release can be relatively big in size.
            raise AirflowSkipException("Skipped, because first release")

        for transform_path in release.transform_files:
            if self.batch_load:
                transform_blob = batch_blob_name(release.transform_folder)
                main_table_id, partition_table_id = self.dag_id, f"{self.dag_id}_partitions"
            else:
                transform_blob = blob_name(transform_path)
                main_table_id, partition_table_id = table_ids_from_path(transform_path)
            table_description = self.table_descriptions.get(main_table_id, "")
            bq_load_ingestion_partition(
                self.schema_folder,
                release.end_date,
                transform_blob,
                self.dataset_id,
                main_table_id,
                partition_table_id,
                self.source_format,
                self.schema_prefix,
                self.schema_version,
                self.dataset_description,
                table_description=table_description,
                **self.load_bigquery_table_kwargs,
            )
            if self.batch_load:
                return

    def bq_delete_old(self, release: StreamRelease, **kwargs):
        """Delete old rows from the 'main' table, based on rows that are in a partition of the 'partitions' table.
        :param release: a StreamRelease instance
        :param kwargs: The context passed from the PythonOperator.
        :return: None.
        """
        ti: TaskInstance = kwargs["ti"]
        if release.first_release:
            # don't use AirflowSkipException, to ensure that task is in 'success' state, this is used below in
            # ti.previous_start_date_success
            logging.info("Skipped, because first release")
            return

        start_date = pendulum.instance(ti.previous_start_date_success)
        end_date = pendulum.instance(ti.start_date)
        if (end_date - start_date).days + 1 >= self.bq_merge_days:
            logging.info(
                f"Deleting old data from main table using partitions after {start_date} and on or before" f" {end_date}"
            )
            for transform_path in release.transform_files:
                if self.batch_load:
                    main_table_id, partition_table_id = self.dag_id, f"{self.dag_id}_partitions"
                    bq_delete_old(
                        start_date,
                        end_date,
                        self.dataset_id,
                        main_table_id,
                        partition_table_id,
                        self.merge_partition_field,
                    )
                    return
                else:
                    main_table_id, partition_table_id = table_ids_from_path(transform_path)
                    bq_delete_old(
                        start_date,
                        end_date,
                        self.dataset_id,
                        main_table_id,
                        partition_table_id,
                        self.merge_partition_field,
                    )
        else:
            raise AirflowSkipException(
                f"Skipped, only delete old records every {self.bq_merge_days} days. "
                f"Last delete was {(end_date - start_date).days + 1} days ago on {ti.previous_start_date_success}"
            )

    def bq_append_new(self, release: StreamRelease, **kwargs):
        """Append rows to the 'main' table, based on rows that are in a partition of the 'partitions' table.
        :param release: a StreamRelease instance
        :param kwargs: The context passed from the PythonOperator.
        :return: None.
        """
        ti: TaskInstance = kwargs["ti"]

        if release.first_release:
            for transform_path in release.transform_files:
                if self.batch_load:
                    transform_blob = batch_blob_name(release.transform_folder)
                    main_table_id, partition_table_id = self.dag_id, f"{self.dag_id}_partitions"
                else:
                    transform_blob = blob_name(transform_path)
                    main_table_id, partition_table_id = table_ids_from_path(transform_path)
                table_description = self.table_descriptions.get(main_table_id, "")
                bq_append_from_file(
                    self.schema_folder,
                    release.end_date,
                    transform_blob,
                    self.dataset_id,
                    main_table_id,
                    self.source_format,
                    self.schema_prefix,
                    self.schema_version,
                    self.dataset_description,
                    table_description=table_description,
                    **self.load_bigquery_table_kwargs,
                )
                if self.batch_load:
                    return
            return

        start_date = pendulum.instance(ti.previous_start_date_success)
        end_date = pendulum.instance(ti.start_date)
        if (end_date - start_date).days + 1 >= self.bq_merge_days:
            logging.info(f"Appending data to main table from partitions after {start_date} and on or before {end_date}")
            for transform_path in release.transform_files:
                if self.batch_load:
                    main_table_id, partition_table_id = self.dag_id, f"{self.dag_id}_partitions"
                    bq_append_from_partition(
                        self.schema_folder,
                        start_date,
                        end_date,
                        self.dataset_id,
                        main_table_id,
                        partition_table_id,
                        self.schema_prefix,
                        self.schema_version,
                    )
                    return
                else:
                    main_table_id, partition_table_id = table_ids_from_path(transform_path)
                    bq_append_from_partition(
                        self.schema_folder,
                        start_date,
                        end_date,
                        self.dataset_id,
                        main_table_id,
                        partition_table_id,
                        self.schema_prefix,
                        self.schema_version,
                    )
        else:
            raise AirflowSkipException(
                f"Skipped, not first release and only append new records every "
                f"{self.bq_merge_days} days. Last append was {(end_date - start_date).days + 1} "
                f"days ago on {ti.previous_start_date_success}"
            )

    def cleanup(self, release: StreamRelease, **kwargs):
        """Delete downloaded, extracted and transformed files of the release.
        :param release: a StreamRelease instance
        :param kwargs: The context passed from the PythonOperator.
        :return: None.
        """
        release.cleanup()



# COPYRIGHT
#
# All contributions by François Chollet:
# Copyright (c) 2015 - 2019, François Chollet.
# All rights reserved.
#
# All contributions by Google:
# Copyright (c) 2015 - 2019, Google, Inc.
# All rights reserved.
#
# All contributions by Microsoft:
# Copyright (c) 2017 - 2019, Microsoft, Inc.
# All rights reserved.
#
# All other contributions:
# Copyright (c) 2015 - 2019, the respective contributors.
# All rights reserved.
#
# Each contributor holds copyright over their respective contributions.
# The project versioning (Git) records all such contribution source information.
#
# LICENSE
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# The contents of this file originate from: https://github.com/keras-team/keras/blob/master/keras/utils/generic_utils.py

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

import numpy as np
import sys
import time

_GLOBAL_CUSTOM_OBJECTS = {}


class Progbar(object):
    """Displays a progress bar.
    # Arguments
        target: Total number of steps expected, None if unknown.
        width: Progress bar width on screen.
        verbose: Verbosity mode, 0 (silent), 1 (verbose), 2 (semi-verbose)
        stateful_metrics: Iterable of string names of metrics that
            should *not* be averaged over time. Metrics in this list
            will be displayed as-is. All others will be averaged
            by the progbar before display.
        interval: Minimum visual progress update interval (in seconds).
    """

    def __init__(self, target, width=30, verbose=1, interval=0.05, stateful_metrics=None):
        self.target = target
        self.width = width
        self.verbose = verbose
        self.interval = interval
        if stateful_metrics:
            self.stateful_metrics = set(stateful_metrics)
        else:
            self.stateful_metrics = set()

        self._dynamic_display = (hasattr(sys.stdout, "isatty") and sys.stdout.isatty()) or "ipykernel" in sys.modules
        self._total_width = 0
        self._seen_so_far = 0
        self._values = collections.OrderedDict()
        self._start = time.time()
        self._last_update = 0

    def update(self, current, values=None):
        """Updates the progress bar.
        # Arguments
            current: Index of current step.
            values: List of tuples:
                `(name, value_for_last_step)`.
                If `name` is in `stateful_metrics`,
                `value_for_last_step` will be displayed as-is.
                Else, an average of the metric over time will be displayed.
        """
        values = values or []
        for k, v in values:
            if k not in self.stateful_metrics:
                if k not in self._values:
                    self._values[k] = [v * (current - self._seen_so_far), current - self._seen_so_far]
                else:
                    self._values[k][0] += v * (current - self._seen_so_far)
                    self._values[k][1] += current - self._seen_so_far
            else:
                # Stateful metrics output a numeric value.  This representation
                # means "take an average from a single value" but keeps the
                # numeric formatting.
                self._values[k] = [v, 1]
        self._seen_so_far = current

        now = time.time()
        info = " - %.0fs" % (now - self._start)
        if self.verbose == 1:
            if now - self._last_update < self.interval and self.target is not None and current < self.target:
                return

            prev_total_width = self._total_width
            if self._dynamic_display:
                sys.stdout.write("\b" * prev_total_width)
                sys.stdout.write("\r")
            else:
                sys.stdout.write("\n")

            if self.target is not None:
                numdigits = int(np.floor(np.log10(self.target))) + 1
                barstr = "%%%dd/%d [" % (numdigits, self.target)
                bar = barstr % current
                prog = float(current) / self.target
                prog_width = int(self.width * prog)
                if prog_width > 0:
                    bar += "=" * (prog_width - 1)
                    if current < self.target:
                        bar += ">"
                    else:
                        bar += "="
                bar += "." * (self.width - prog_width)
                bar += "]"
            else:
                bar = "%7d/Unknown" % current

            self._total_width = len(bar)
            sys.stdout.write(bar)

            if current:
                time_per_unit = (now - self._start) / current
            else:
                time_per_unit = 0
            if self.target is not None and current < self.target:
                eta = time_per_unit * (self.target - current)
                if eta > 3600:
                    eta_format = "%d:%02d:%02d" % (eta // 3600, (eta % 3600) // 60, eta % 60)
                elif eta > 60:
                    eta_format = "%d:%02d" % (eta // 60, eta % 60)
                else:
                    eta_format = "%ds" % eta

                info = " - ETA: %s" % eta_format
            else:
                if time_per_unit >= 1:
                    info += " %.0fs/step" % time_per_unit
                elif time_per_unit >= 1e-3:
                    info += " %.0fms/step" % (time_per_unit * 1e3)
                else:
                    info += " %.0fus/step" % (time_per_unit * 1e6)

            for k in self._values:
                info += " - %s:" % k
                if isinstance(self._values[k], list):
                    avg = np.mean(self._values[k][0] / max(1, self._values[k][1]))
                    if abs(avg) > 1e-3:
                        info += " %.4f" % avg
                    else:
                        info += " %.4e" % avg
                else:
                    info += " %s" % self._values[k]

            self._total_width += len(info)
            if prev_total_width > self._total_width:
                info += " " * (prev_total_width - self._total_width)

            if self.target is not None and current >= self.target:
                info += "\n"

            sys.stdout.write(info)
            sys.stdout.flush()

        elif self.verbose == 2:
            if self.target is None or current >= self.target:
                for k in self._values:
                    info += " - %s:" % k
                    avg = np.mean(self._values[k][0] / max(1, self._values[k][1]))
                    if avg > 1e-3:
                        info += " %.4f" % avg
                    else:
                        info += " %.4e" % avg
                info += "\n"

                sys.stdout.write(info)
                sys.stdout.flush()

        self._last_update = now

    def add(self, n, values=None):
        self.update(self._seen_so_far + n, values)



# Copyright 2019, 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose, Aniek Roelofs


from __future__ import annotations

import binascii
import json
import os
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, ClassVar, Dict, List, Tuple, Union

import cerberus.validator
import yaml
from cerberus import Validator
from cryptography.fernet import Fernet

from observatory.platform.terraform_api import TerraformVariable
from observatory.platform.utils.airflow_utils import AirflowVars
from observatory.platform.utils.config_utils import module_file_path
from observatory.platform.utils.config_utils import observatory_home as default_observatory_home
from observatory.platform.utils.jinja2_utils import render_template


def generate_secret_key(length: int = 30) -> str:
    """Generate a secret key for the Flask Airflow Webserver.

    :param length: the length of the key to generate.
    :return: the random key.
    """

    return str(binascii.b2a_hex(os.urandom(length)))


def save_yaml(file_path: str, dict_: Dict):
    """Save a yaml file from a dictionary.

    :param file_path: the path to the file to save.
    :param dict_: the dictionary.
    :return: None.
    """

    with open(file_path, "w") as yaml_file:
        yaml.dump(dict_, yaml_file, default_flow_style=False)


def to_hcl(value: Dict) -> str:
    """Convert a Python dictionary into HCL.

    :param value: the dictionary.
    :return: the HCL string.
    """

    return json.dumps(value, separators=(",", "="))


def from_hcl(string: str) -> Dict:
    """Convert an HCL string into a Dict.

    :param string: the HCL string.
    :return: the dict.
    """

    return json.loads(string.replace('"=', '":'))


class BackendType(Enum):
    """The type of backend"""

    local = "local"
    terraform = "terraform"


class Environment(Enum):
    """The environment being used"""

    develop = "develop"
    staging = "staging"
    production = "production"


@dataclass
class Backend:
    """The backend settings for the Observatory Platform.

    Attributes:
        type: the type of backend being used (local environment or Terraform).
        environment: what type of environment is being deployed (develop, staging or production).
    """

    type: BackendType
    environment: Environment

    @staticmethod
    def from_dict(dict_: Dict) -> Backend:
        """Constructs a Backend instance from a dictionary.

        :param dict_: the dictionary.
        :return: the Backend instance.
        """

        backend_type = BackendType(dict_.get("type"))
        environment = Environment(dict_.get("environment"))

        return Backend(
            backend_type,
            environment,
        )


@dataclass
class PythonPackage:
    name: str
    host_package: str
    docker_package: str
    type: str


@dataclass
class Observatory:
    """The Observatory settings for the Observatory Platform.

    Attributes:
        :param package: the observatory platform package, either a local path to a Python source package (editable type),
        path to a sdist (sdist) or a PyPI package name and version (pypi).
        :param package_type: the package type, editable, sdist, pypi.
        :param airflow_fernet_key: the Fernet key.
        :param airflow_secret_key: the secret key used to run the flask app.
        :param airflow_ui_user_password: the password for the Apache Airflow UI admin user.
        :param airflow_ui_user_email: the email address for the Apache Airflow UI admin user.
        :param observatory_home: The observatory home folder.
        :param postgres_password: the Postgres SQL password.
        :param redis_port: The host Redis port number.
        :param flower_ui_port: The host's Flower UI port number.
        :param airflow_ui_port: The host's Apache Airflow UI port number.
        :param elastic_port: The host's Elasticsearch port number.
        :param kibana_port: The host's Kibana port number.
        :param docker_network_name: The Docker Network name, used to specify a custom Docker Network.
        :param docker_network_is_external: whether the docker network is external or not.
        :param docker_compose_project_name: The namespace for the Docker Compose containers: https://docs.docker.com/compose/reference/envvars/#compose_project_name.
        :param enable_elk: whether to enable the elk stack or not.
    """

    package: str
    package_type: str
    airflow_fernet_key: str
    airflow_secret_key: str
    airflow_ui_user_email: str = "airflow@airflow.com"
    airflow_ui_user_password: str = "airflow"
    observatory_home: str = default_observatory_home()
    postgres_password: str = "postgres"
    redis_port: int = 6379
    flower_ui_port: int = 5555
    airflow_ui_port: int = 8080
    elastic_port: int = 9200
    kibana_port: int = 5601
    docker_network_name: str = "observatory-network"
    docker_network_is_external: bool = False
    docker_compose_project_name: str = "observatory"
    enable_elk: bool = True
    api_package: str = "observatory-api"
    api_package_type: str = "pypi"

    def to_hcl(self):
        return to_hcl(
            {
                "airflow_fernet_key": self.airflow_fernet_key,
                "airflow_secret_key": self.airflow_secret_key,
                "airflow_ui_user_password": self.airflow_ui_user_password,
                "airflow_ui_user_email": self.airflow_ui_user_email,
                "postgres_password": self.postgres_password,
            }
        )

    @property
    def host_package(self):
        return os.path.normpath(self.package)

    @staticmethod
    def from_dict(dict_: Dict) -> Observatory:
        """Constructs an Airflow instance from a dictionary.

        :param dict_: the dictionary.
        :return: the Airflow instance.
        """

        package = dict_.get("package")
        package_type = dict_.get("package_type")
        airflow_fernet_key = dict_.get("airflow_fernet_key")
        airflow_secret_key = dict_.get("airflow_secret_key")
        airflow_ui_user_email = dict_.get("airflow_ui_user_email", Observatory.airflow_ui_user_email)
        airflow_ui_user_password = dict_.get("airflow_ui_user_password", Observatory.airflow_ui_user_password)
        observatory_home = dict_.get("observatory_home", Observatory.observatory_home)
        postgres_password = dict_.get("postgres_password", Observatory.postgres_password)
        redis_port = dict_.get("redis_port", Observatory.redis_port)
        flower_ui_port = dict_.get("flower_ui_port", Observatory.flower_ui_port)
        airflow_ui_port = dict_.get("airflow_ui_port", Observatory.airflow_ui_port)
        elastic_port = dict_.get("elastic_port", Observatory.elastic_port)
        kibana_port = dict_.get("kibana_port", Observatory.kibana_port)
        docker_network_name = dict_.get("docker_network_name", Observatory.docker_network_name)
        docker_network_is_external = dict_.get("docker_network_is_external", Observatory.docker_network_is_external)
        docker_compose_project_name = dict_.get("docker_compose_project_name", Observatory.docker_compose_project_name)
        enable_elk = dict_.get("enable_elk", Observatory.enable_elk)
        api_package = dict_.get("api_package", Observatory.api_package)
        api_package_type = dict_.get("api_package_type", Observatory.api_package_type)

        return Observatory(
            package,
            package_type,
            airflow_fernet_key,
            airflow_secret_key,
            airflow_ui_user_password=airflow_ui_user_password,
            airflow_ui_user_email=airflow_ui_user_email,
            observatory_home=observatory_home,
            postgres_password=postgres_password,
            redis_port=redis_port,
            flower_ui_port=flower_ui_port,
            airflow_ui_port=airflow_ui_port,
            elastic_port=elastic_port,
            kibana_port=kibana_port,
            docker_network_name=docker_network_name,
            docker_network_is_external=docker_network_is_external,
            docker_compose_project_name=docker_compose_project_name,
            enable_elk=enable_elk,
            api_package=api_package,
            api_package_type=api_package_type,
        )


@dataclass
class CloudStorageBucket:
    """Represents a Google Cloud Storage Bucket.

    Attributes:
        id: the id of the bucket (which gets set as an Airflow variable).
        name: the name of the Google Cloud storage bucket.
    """

    id: str
    name: str

    @staticmethod
    def parse_buckets(buckets: Dict) -> List[CloudStorageBucket]:
        return parse_dict_to_list(buckets, CloudStorageBucket)


@dataclass
class GoogleCloud:
    """The Google Cloud settings for the Observatory Platform.

    Attributes:
        project_id: the Google Cloud project id.
        credentials: the path to the Google Cloud credentials.
        region: the Google Cloud region.
        zone: the Google Cloud zone.
        data_location: the data location for storing buckets.
        buckets: a list of the Google Cloud buckets.
    """

    project_id: str = None
    credentials: str = None
    region: str = None
    zone: str = None
    data_location: str = None
    buckets: List[CloudStorageBucket] = field(default_factory=lambda: [])

    def read_credentials(self) -> str:
        with open(self.credentials, "r") as f:
            data = f.read()
        return data

    def to_hcl(self):
        return to_hcl(
            {
                "project_id": self.project_id,
                "credentials": self.read_credentials(),
                "region": self.region,
                "zone": self.zone,
                "data_location": self.data_location,
                "buckets": [bucket.name for bucket in self.buckets],
            }
        )

    @staticmethod
    def from_dict(dict_: Dict) -> GoogleCloud:
        """Constructs a GoogleCloud instance from a dictionary.

        :param dict_: the dictionary.
        :return: the GoogleCloud instance.
        """

        project_id = dict_.get("project_id")
        credentials = dict_.get("credentials")
        region = dict_.get("region")
        zone = dict_.get("zone")
        data_location = dict_.get("data_location")
        buckets = CloudStorageBucket.parse_buckets(dict_.get("buckets", dict()))

        return GoogleCloud(
            project_id=project_id,
            credentials=credentials,
            region=region,
            zone=zone,
            data_location=data_location,
            buckets=buckets,
        )


def parse_dict_to_list(dict_: Dict, cls: ClassVar) -> List[Any]:
    """Parse the key, value pairs in a dictionary into a list of class instances.

    :param dict_: the dictionary.
    :param cls: the type of class to construct.
    :return: a list of class instances.
    """

    parsed_items = []
    for key, val in dict_.items():
        parsed_items.append(cls(key, val))
    return parsed_items


@dataclass
class WorkflowsProject:
    """Represents a project that contains DAGs to load.

    Attributes:
        :param package_name: the package name.
        :param package: the observatory platform package, either a local path to a Python source package (editable type),
        path to a sdist (sdist) or a PyPI package name and version (pypi).
        :param package_type: the package type, editable, sdist, pypi.
        dags_module: the Python import path to the module that contains the Apache Airflow DAGs to load.
    """

    package_name: str
    package: str
    package_type: str
    dags_module: str

    @staticmethod
    def parse_dags_projects(list: List) -> List[WorkflowsProject]:
        """Parse the workflows_projects list object into a list of WorkflowsProject instances.

        :param list: the list.
        :return: a list of WorkflowsProject instances.
        """

        parsed_items = []
        for item in list:
            package_name = item["package_name"]
            package = item["package"]
            package_type = item["package_type"]
            dags_module = item["dags_module"]
            parsed_items.append(WorkflowsProject(package_name, package, package_type, dags_module))
        return parsed_items


def list_to_hcl(items: List[Union[AirflowConnection, AirflowVariable]]) -> str:
    """Convert a list of AirflowConnection or AirflowVariable instances into HCL.

    :param items: the list of AirflowConnection or AirflowVariable instances.
    :return: the HCL string.
    """

    dict_ = dict()
    for item in items:
        dict_[item.name] = item.value
    return to_hcl(dict_)


@dataclass
class AirflowConnection:
    """An Airflow Connection.

    Attributes:
        name: the name of the Airflow Connection.
        value: the value of the Airflow Connection.
    """

    name: str
    value: str

    @property
    def conn_name(self) -> str:
        """The Airflow Connection environment variable name, which is required to set the connection from an
        environment variable.

        :return: the Airflow Connection environment variable name.
        """

        return f"AIRFLOW_CONN_{self.name.upper()}"

    @staticmethod
    def parse_airflow_connections(dict_: Dict) -> List[AirflowConnection]:
        """Parse the airflow_connections dictionary object into a list of AirflowConnection instances.

        :param dict_: the dictionary.
        :return: a list of AirflowConnection instances.
        """

        return parse_dict_to_list(dict_, AirflowConnection)


@dataclass
class AirflowVariable:
    """An Airflow Variable.

    Attributes:
        name: the name of the Airflow Variable.
        value: the value of the Airflow Variable.
    """

    name: str
    value: str

    @property
    def env_var_name(self):
        """The Airflow Variable environment variable name, which is required to set the variable from an
        environment variable.

        :return: the Airflow Variable environment variable name.
        """

        return f"AIRFLOW_VAR_{self.name.upper()}"

    @staticmethod
    def parse_airflow_variables(dict_: Dict) -> List[AirflowVariable]:
        """Parse the airflow_variables dictionary object into a list of AirflowVariable instances.

        :param dict_: the dictionary.
        :return: a list of AirflowVariable instances.
        """

        return parse_dict_to_list(dict_, AirflowVariable)


@dataclass
class Terraform:
    """The Terraform settings for the Observatory Platform.

    Attributes:
        organization: the Terraform Organisation name.
    """

    organization: str

    @staticmethod
    def from_dict(dict_: Dict) -> Terraform:
        """Constructs a Terraform instance from a dictionary.

        :param dict_: the dictionary.
        """

        organization = dict_.get("organization")
        return Terraform(organization)


@dataclass
class CloudSqlDatabase:
    """The Google Cloud SQL database settings for the Observatory Platform.

    Attributes:
        tier: the database machine tier.
        backup_start_time: the start time for backups in HH:MM format.
    """

    tier: str
    backup_start_time: str

    def to_hcl(self):
        return to_hcl({"tier": self.tier, "backup_start_time": self.backup_start_time})

    @staticmethod
    def from_dict(dict_: Dict) -> CloudSqlDatabase:
        """Constructs a CloudSqlDatabase instance from a dictionary.

        :param dict_: the dictionary.
        :return: the CloudSqlDatabase instance.
        """

        tier = dict_.get("tier")
        backup_start_time = dict_.get("backup_start_time")
        return CloudSqlDatabase(tier, backup_start_time)


@dataclass
class VirtualMachine:
    """A Google Cloud virtual machine.

    Attributes:
        machine_type: the type of Google Cloud virtual machine.
        disk_size: the size of the disk in GB.
        disk_type: the disk type; pd-standard or pd-ssd.
        create: whether to create the VM or not.
    """

    machine_type: str
    disk_size: int
    disk_type: str
    create: bool

    def to_hcl(self):
        return to_hcl(
            {
                "machine_type": self.machine_type,
                "disk_size": self.disk_size,
                "disk_type": self.disk_type,
                "create": self.create,
            }
        )

    @staticmethod
    def from_hcl(string: str) -> VirtualMachine:
        return VirtualMachine.from_dict(from_hcl(string))

    @staticmethod
    def from_dict(dict_: Dict) -> VirtualMachine:
        """Constructs a VirtualMachine instance from a dictionary.

        :param dict_: the dictionary.
        :return: the VirtualMachine instance.
        """

        machine_type = dict_.get("machine_type")
        disk_size = dict_.get("disk_size")
        disk_type = dict_.get("disk_type")
        create = str(dict_.get("create")).lower() == "true"
        return VirtualMachine(machine_type, disk_size, disk_type, create)


@dataclass
class ElasticSearch:
    """The elasticsearch settings for the Observatory Platform API.

    Attributes:
        host: the address of the elasticsearch host
        api_key: the api key to use the elasticsearch API.
    """

    host: str
    api_key: str

    def to_hcl(self):
        return to_hcl(
            {
                "host": self.host,
                "api_key": self.api_key,
            }
        )

    @staticmethod
    def from_dict(dict_: Dict) -> ElasticSearch:
        """Constructs a CloudSqlDatabase instance from a dictionary.

        :param dict_: the dictionary.
        :return: the CloudSqlDatabase instance.
        """

        host = dict_.get("host")
        api_key = dict_.get("api_key")
        return ElasticSearch(host, api_key)


@dataclass
class Api:
    """The API domain name for the Observatory Platform API.

    Attributes:
        domain_name: the custom domain name of the API
        subdomain: the subdomain of the API, can be either based on the google project id or the environment. When
        based on the environment, there is no subdomain for the production environment.
    """

    domain_name: str
    subdomain: str

    def to_hcl(self):
        return to_hcl({"domain_name": self.domain_name, "subdomain": self.subdomain})

    @staticmethod
    def from_dict(dict_: Dict) -> Api:
        """Constructs a CloudSqlDatabase instance from a dictionary.

        :param dict_: the dictionary.
        :return: the CloudSqlDatabase instance.
        """

        domain_name = dict_.get("domain_name")
        subdomain = dict_.get("subdomain")
        return Api(domain_name, subdomain)


def customise_pointer(field, value, error):
    """Throw an error when a field contains the value ' <--' which means that the user should customise the
    value in the config file.

    :param field: the field.
    :param value: the value.
    :param error: ?
    :return: None.
    """

    if isinstance(value, str) and value.endswith(" <--"):
        error(field, "Customise value ending with ' <--'")


class ObservatoryConfigValidator(Validator):
    """Custom config Validator"""

    def _validate_google_application_credentials(self, google_application_credentials, field, value):
        """Validate that the Google Application Credentials file exists.
        The rule's arguments are validated against this schema: {'type': 'boolean'}
        """
        if (
            google_application_credentials
            and value is not None
            and isinstance(value, str)
            and not os.path.isfile(value)
        ):
            self._error(
                field,
                f"the file {value} does not exist. See "
                f"https://cloud.google.com/docs/authentication/getting-started for instructions on "
                f"how to create a service account and save the JSON key to your workstation.",
            )


@dataclass
class ValidationError:
    """A validation error found when parsing a config file.

    Attributes:
        key: the key in the config file.
        value: the error.
    """

    key: str
    value: Any


class ObservatoryConfig:
    def __init__(
        self,
        backend: Backend = None,
        observatory: Observatory = None,
        google_cloud: GoogleCloud = None,
        terraform: Terraform = None,
        airflow_variables: List[AirflowVariable] = None,
        airflow_connections: List[AirflowConnection] = None,
        workflows_projects: List[WorkflowsProject] = None,
        validator: ObservatoryConfigValidator = None,
    ):
        """Create an ObservatoryConfig instance.

        :param backend: the backend config.
        :param observatory: the Observatory config.
        :param google_cloud: the Google Cloud config.
        :param terraform: the Terraform config.
        :param airflow_variables: a list of Airflow variables.
        :param airflow_connections: a list of Airflow connections.
        :param workflows_projects: a list of DAGs projects.
        :param validator: an ObservatoryConfigValidator instance.
        """

        self.backend = backend
        self.observatory = observatory
        self.google_cloud = google_cloud
        self.terraform = terraform

        self.airflow_variables = airflow_variables
        if airflow_variables is None:
            self.airflow_variables = []

        self.airflow_connections = airflow_connections
        if airflow_variables is None:
            self.airflow_connections = []

        self.workflows_projects = workflows_projects
        if workflows_projects is None:
            self.workflows_projects = []

        self.validator = validator

    @property
    def is_valid(self) -> bool:
        """Checks whether the config is valid or not.

        :return: whether the config is valid or not.
        """

        return self.validator is None or not len(self.validator._errors)

    @property
    def errors(self) -> List[ValidationError]:
        """Returns a list of ValidationError instances that were created when parsing the config file.

        :return: the list of ValidationError instances.
        """

        errors = []
        for key, values in self.validator.errors.items():
            for value in values:
                if type(value) is dict:
                    for nested_key, nested_value in value.items():
                        errors.append(ValidationError(f"{key}.{nested_key}", *nested_value))
                else:
                    errors.append(ValidationError(key, *values))

        return errors

    @property
    def python_packages(self) -> List[PythonPackage]:
        """Make a list of Python Packages to build or include in the observatory.
        :return: the list of Python packages.
        """

        packages = [
            PythonPackage(
                name="observatory-api",
                type=self.observatory.api_package_type,
                host_package=self.observatory.api_package,
                docker_package=os.path.basename(self.observatory.api_package),
            ),
            PythonPackage(
                name="observatory-platform",
                type=self.observatory.package_type,
                host_package=self.observatory.package,
                docker_package=os.path.basename(self.observatory.package),
            ),
        ]

        for project in self.workflows_projects:
            packages.append(
                PythonPackage(
                    name=project.package_name,
                    type=project.package_type,
                    host_package=project.package,
                    docker_package=os.path.basename(project.package),
                )
            )

        return packages

    @property
    def dags_module_names(self):
        """Returns a list of DAG project module names.
        :return: the list of DAG project module names.
        """

        return f"'{str(json.dumps([project.dags_module for project in self.workflows_projects]))}'"

    def make_airflow_variables(self) -> List[AirflowVariable]:
        """Make all airflow variables for the observatory platform.

        :return: a list of AirflowVariable objects.
        """

        # Create airflow variables from fixed config file values
        variables = [AirflowVariable(AirflowVars.ENVIRONMENT, self.backend.environment.value)]

        if self.google_cloud is not None:
            if self.google_cloud.project_id is not None:
                variables.append(AirflowVariable(AirflowVars.PROJECT_ID, self.google_cloud.project_id))

            if self.google_cloud.data_location:
                variables.append(AirflowVariable(AirflowVars.DATA_LOCATION, self.google_cloud.data_location))

            # Create airflow variables from bucket names
            for bucket in self.google_cloud.buckets:
                variables.append(AirflowVariable(bucket.id, bucket.name))

        if self.terraform is not None:
            if self.terraform.organization is not None:
                variables.append(AirflowVariable(AirflowVars.TERRAFORM_ORGANIZATION, self.terraform.organization))

        # Add user defined variables to list
        variables += self.airflow_variables

        return variables

    @staticmethod
    def _parse_fields(
        dict_: Dict,
    ) -> Tuple[
        Backend,
        Observatory,
        GoogleCloud,
        Terraform,
        List[AirflowVariable],
        List[AirflowConnection],
        List[WorkflowsProject],
    ]:
        backend = Backend.from_dict(dict_.get("backend", dict()))
        observatory = Observatory.from_dict(dict_.get("observatory", dict()))
        google_cloud = GoogleCloud.from_dict(dict_.get("google_cloud", dict()))
        terraform = Terraform.from_dict(dict_.get("terraform", dict()))
        airflow_variables = AirflowVariable.parse_airflow_variables(dict_.get("airflow_variables", dict()))
        airflow_connections = AirflowConnection.parse_airflow_connections(dict_.get("airflow_connections", dict()))
        workflows_projects = WorkflowsProject.parse_dags_projects(dict_.get("workflows_projects", list()))

        return backend, observatory, google_cloud, terraform, airflow_variables, airflow_connections, workflows_projects

    @staticmethod
    def save_default(config_path: str):
        """Save a default config file.

        :param config_path: the path where the config file should be saved.
        :return: None.
        """

        ObservatoryConfig._save_default(config_path, "config.yaml.jinja2")

    @staticmethod
    def _save_default(config_path: str, template_file_name: str):
        """Save a default config file.

        :param config_path: the path where the config file should be saved.
        :param template_file_name: the name of the template file name.
        :return: None.
        """

        # Render template
        template_path = os.path.join(module_file_path("observatory.platform"), template_file_name)
        airflow_fernet_key = Fernet.generate_key()
        airflow_secret_key = generate_secret_key()

        render = render_template(
            template_path, airflow_fernet_key=airflow_fernet_key, airflow_secret_key=airflow_secret_key
        )

        # Save file
        with open(config_path, "w") as f:
            f.write(render)

    @classmethod
    def from_dict(cls, dict_: Dict) -> ObservatoryConfig:
        """Constructs an ObservatoryConfig instance from a dictionary.

        If the dictionary is invalid, then an ObservatoryConfig instance will be returned with no properties set,
        except for the validator, which contains validation errors.

        :param dict_: the input dictionary.
        :return: the ObservatoryConfig instance.
        """

        schema = make_schema(BackendType.local)
        validator = ObservatoryConfigValidator()
        is_valid = validator.validate(dict_, schema)

        if is_valid:
            (
                backend,
                observatory,
                google_cloud,
                terraform,
                airflow_variables,
                airflow_connections,
                workflows_projects,
            ) = ObservatoryConfig._parse_fields(dict_)

            return ObservatoryConfig(
                backend,
                observatory,
                google_cloud=google_cloud,
                terraform=terraform,
                airflow_variables=airflow_variables,
                airflow_connections=airflow_connections,
                workflows_projects=workflows_projects,
                validator=validator,
            )
        else:
            return ObservatoryConfig(validator=validator)

    @classmethod
    def load(cls, path: str):
        """Load a configuration file.

        :return: the ObservatoryConfig instance (or a subclass of ObservatoryConfig)
        """

        dict_ = dict()

        try:
            with open(path, "r") as f:
                dict_ = yaml.safe_load(f)
        except yaml.YAMLError:
            print(f"Error parsing {path}")
        except FileNotFoundError:
            print(f"No such file or directory: {path}")
        except cerberus.validator.DocumentError as e:
            print(f"cerberus.validator.DocumentError: {e}")

        return cls.from_dict(dict_)


class TerraformConfig(ObservatoryConfig):
    WORKSPACE_PREFIX = "observatory-"

    def __init__(
        self,
        backend: Backend = None,
        observatory: Observatory = None,
        google_cloud: GoogleCloud = None,
        terraform: Terraform = None,
        airflow_variables: List[AirflowVariable] = None,
        airflow_connections: List[AirflowConnection] = None,
        workflows_projects: List[WorkflowsProject] = None,
        cloud_sql_database: CloudSqlDatabase = None,
        airflow_main_vm: VirtualMachine = None,
        airflow_worker_vm: VirtualMachine = None,
        elasticsearch: ElasticSearch = None,
        api: Api = None,
        validator: ObservatoryConfigValidator = None,
    ):
        """Create a TerraformConfig instance.

        :param backend: the backend config.
        :param observatory: the Observatory config.
        :param google_cloud: the Google Cloud config.
        :param terraform: the Terraform config.
        :param airflow_variables: a list of Airflow variables.
        :param airflow_connections: a list of Airflow connections.
        :param workflows_projects: a list of DAGs projects.
        :param cloud_sql_database: a Google Cloud SQL database config.
        :param airflow_main_vm: the Airflow Main VM config.
        :param airflow_worker_vm: the Airflow Worker VM config.
        :param validator: an ObservatoryConfigValidator instance.
        """

        super().__init__(
            backend=backend,
            observatory=observatory,
            google_cloud=google_cloud,
            terraform=terraform,
            airflow_variables=airflow_variables,
            airflow_connections=airflow_connections,
            workflows_projects=workflows_projects,
            validator=validator,
        )
        self.cloud_sql_database = cloud_sql_database
        self.airflow_main_vm = airflow_main_vm
        self.airflow_worker_vm = airflow_worker_vm
        self.elasticsearch = elasticsearch
        self.api = api

    @property
    def terraform_workspace_id(self):
        """The Terraform workspace id.

        :return: the terraform workspace id.
        """

        return TerraformConfig.WORKSPACE_PREFIX + self.backend.environment.value

    def make_airflow_variables(self) -> List[AirflowVariable]:
        """Make all airflow variables for the Observatory Platform.

        :return: a list of AirflowVariable objects.
        """

        # Create airflow variables from fixed config file values
        variables = [AirflowVariable(AirflowVars.ENVIRONMENT, self.backend.environment.value)]

        if self.google_cloud.project_id is not None:
            variables.append(AirflowVariable(AirflowVars.PROJECT_ID, self.google_cloud.project_id))
            variables.append(AirflowVariable(AirflowVars.DOWNLOAD_BUCKET, f"{self.google_cloud.project_id}-download"))
            variables.append(AirflowVariable(AirflowVars.TRANSFORM_BUCKET, f"{self.google_cloud.project_id}-transform"))

        if self.google_cloud.data_location:
            variables.append(AirflowVariable(AirflowVars.DATA_LOCATION, self.google_cloud.data_location))

        if self.terraform.organization is not None:
            variables.append(AirflowVariable(AirflowVars.TERRAFORM_ORGANIZATION, self.terraform.organization))

        variables.append(
            AirflowVariable(
                AirflowVars.DAGS_MODULE_NAMES, json.dumps([proj.dags_module for proj in self.workflows_projects])
            )
        )

        # Add user defined variables to list
        variables += self.airflow_variables

        return variables

    def terraform_variables(self) -> List[TerraformVariable]:
        """Create a list of TerraformVariable instances from the Terraform Config.

        :return: a list of TerraformVariable instances.
        """

        sensitive = True
        return [
            TerraformVariable("environment", self.backend.environment.value),
            TerraformVariable("observatory", self.observatory.to_hcl(), sensitive=sensitive, hcl=True),
            TerraformVariable("google_cloud", self.google_cloud.to_hcl(), sensitive=sensitive, hcl=True),
            TerraformVariable("cloud_sql_database", self.cloud_sql_database.to_hcl(), hcl=True),
            TerraformVariable("airflow_main_vm", self.airflow_main_vm.to_hcl(), hcl=True),
            TerraformVariable("airflow_worker_vm", self.airflow_worker_vm.to_hcl(), hcl=True),
            TerraformVariable(
                "airflow_variables", list_to_hcl(self.make_airflow_variables()), hcl=True, sensitive=False
            ),
            TerraformVariable(
                "airflow_connections", list_to_hcl(self.airflow_connections), hcl=True, sensitive=sensitive
            ),
            TerraformVariable("elasticsearch", self.elasticsearch.to_hcl(), sensitive=sensitive, hcl=True),
            TerraformVariable("api", self.api.to_hcl(), hcl=True),
        ]

    @classmethod
    def from_dict(cls, dict_: Dict) -> TerraformConfig:
        """Make an TerraformConfig instance from a dictionary.

        If the dictionary is invalid, then an ObservatoryConfig instance will be returned with no properties set,
        except for the validator, which contains validation errors.

        :param dict_: the input dictionary that has been read via yaml.safe_load.
        :return: the TerraformConfig instance.
        """

        schema = make_schema(BackendType.terraform)
        validator = ObservatoryConfigValidator()
        is_valid = validator.validate(dict_, schema)

        if is_valid:
            (
                backend,
                observatory,
                google_cloud,
                terraform,
                airflow_variables,
                airflow_connections,
                workflows_projects,
            ) = ObservatoryConfig._parse_fields(dict_)

            cloud_sql_database = CloudSqlDatabase.from_dict(dict_.get("cloud_sql_database", dict()))
            airflow_main_vm = VirtualMachine.from_dict(dict_.get("airflow_main_vm", dict()))
            airflow_worker_vm = VirtualMachine.from_dict(dict_.get("airflow_worker_vm", dict()))
            elasticsearch = ElasticSearch.from_dict(dict_.get("elasticsearch", dict()))
            api = Api.from_dict(dict_.get("api", dict()))

            return TerraformConfig(
                backend,
                observatory,
                google_cloud=google_cloud,
                terraform=terraform,
                airflow_variables=airflow_variables,
                airflow_connections=airflow_connections,
                workflows_projects=workflows_projects,
                cloud_sql_database=cloud_sql_database,
                airflow_main_vm=airflow_main_vm,
                airflow_worker_vm=airflow_worker_vm,
                elasticsearch=elasticsearch,
                api=api,
                validator=validator,
            )
        else:
            return TerraformConfig(validator=validator)

    @staticmethod
    def save_default(config_path: str):
        """Save a default TerraformConfig file.

        :param config_path: the path where the config file should be saved.
        :return: None.
        """

        ObservatoryConfig._save_default(config_path, "config-terraform.yaml.jinja2")


def make_schema(backend_type: BackendType) -> Dict:
    """Make a schema for an Observatory or Terraform config file.

    :param backend_type: the type of backend, local or terraform.
    :return: a dictionary containing the schema.
    """

    schema = dict()
    is_backend_terraform = backend_type == BackendType.terraform

    # Backend settings
    schema["backend"] = {
        "required": True,
        "type": "dict",
        "schema": {
            "type": {"required": True, "type": "string", "allowed": [backend_type.value]},
            "environment": {"required": True, "type": "string", "allowed": ["develop", "staging", "production"]},
        },
    }

    # Terraform settings
    schema["terraform"] = {
        "required": is_backend_terraform,
        "type": "dict",
        "schema": {"organization": {"required": True, "type": "string", "check_with": customise_pointer}},
    }

    # Google Cloud settings
    schema["google_cloud"] = {
        "required": is_backend_terraform,
        "type": "dict",
        "schema": {
            "project_id": {"required": is_backend_terraform, "type": "string", "check_with": customise_pointer},
            "credentials": {
                "required": is_backend_terraform,
                "type": "string",
                "check_with": customise_pointer,
                "google_application_credentials": True,
            },
            "region": {
                "required": is_backend_terraform,
                "type": "string",
                "regex": r"^\w+\-\w+\d+$",
                "check_with": customise_pointer,
            },
            "zone": {
                "required": is_backend_terraform,
                "type": "string",
                "regex": r"^\w+\-\w+\d+\-[a-z]{1}$",
                "check_with": customise_pointer,
            },
            "data_location": {"required": is_backend_terraform, "type": "string", "check_with": customise_pointer},
        },
    }

    if not is_backend_terraform:
        schema["google_cloud"]["schema"]["buckets"] = {
            "required": False,
            "type": "dict",
            "keysrules": {"type": "string"},
            "valuesrules": {"type": "string"},
        }

    # Observatory settings
    package_types = ["editable", "sdist", "pypi"]
    schema["observatory"] = {
        "required": True,
        "type": "dict",
        "schema": {
            "package": {"required": True, "type": "string"},
            "package_type": {"required": True, "type": "string", "allowed": package_types},
            "airflow_fernet_key": {"required": True, "type": "string"},
            "airflow_secret_key": {"required": True, "type": "string"},
            "airflow_ui_user_password": {"required": is_backend_terraform, "type": "string"},
            "airflow_ui_user_email": {"required": is_backend_terraform, "type": "string"},
            "observatory_home": {"required": False, "type": "string"},
            "postgres_password": {"required": is_backend_terraform, "type": "string"},
            "redis_port": {"required": False, "type": "integer"},
            "flower_ui_port": {"required": False, "type": "integer"},
            "airflow_ui_port": {"required": False, "type": "integer"},
            "elastic_port": {"required": False, "type": "integer"},
            "kibana_port": {"required": False, "type": "integer"},
            "docker_network_name": {"required": False, "type": "string"},
            "docker_network_is_external": {"required": False, "type": "boolean"},
            "docker_compose_project_name": {"required": False, "type": "string"},
            "enable_elk": {"required": False, "type": "boolean"},
            "api_package": {"required": False, "type": "string"},
            "api_package_type": {"required": False, "type": "string", "allowed": package_types},
        },
    }

    # Database settings
    if is_backend_terraform:
        schema["cloud_sql_database"] = {
            "required": True,
            "type": "dict",
            "schema": {
                "tier": {"required": True, "type": "string"},
                "backup_start_time": {"required": True, "type": "string", "regex": r"^\d{2}:\d{2}$"},
            },
        }

    # VM schema
    vm_schema = {
        "required": True,
        "type": "dict",
        "schema": {
            "machine_type": {
                "required": True,
                "type": "string",
            },
            "disk_size": {"required": True, "type": "integer", "min": 1},
            "disk_type": {"required": True, "type": "string", "allowed": ["pd-standard", "pd-ssd"]},
            "create": {"required": True, "type": "boolean"},
        },
    }

    # Airflow main and worker VM
    if is_backend_terraform:
        schema["airflow_main_vm"] = vm_schema
        schema["airflow_worker_vm"] = vm_schema

    # Key value string pair schema
    key_val_schema = {
        "required": False,
        "type": "dict",
        "keysrules": {"type": "string"},
        "valuesrules": {"type": "string"},
    }

    # Airflow variables
    schema["airflow_variables"] = key_val_schema

    # Airflow connections
    schema["airflow_connections"] = {
        "required": False,
        "type": "dict",
        "keysrules": {"type": "string"},
        "valuesrules": {"type": "string", "regex": r"\S*:\/\/\S*:\S*@\S*$|google-cloud-platform:\/\/\S*$"},
    }

    # Dags projects
    schema["workflows_projects"] = {
        "required": False,
        "type": "list",
        "schema": {
            "type": "dict",
            "schema": {
                "package_name": {
                    "required": True,
                    "type": "string",
                },
                "package": {"required": True, "type": "string"},
                "package_type": {"required": True, "type": "string", "allowed": package_types},
                "dags_module": {
                    "required": True,
                    "type": "string",
                },
            },
        },
    }

    if is_backend_terraform:
        schema["elasticsearch"] = {
            "required": True,
            "type": "dict",
            "schema": {
                "host": {"required": True, "type": "string"},
                "api_key": {
                    "required": True,
                    "type": "string",
                },
            },
        }

        schema["api"] = {
            "required": True,
            "type": "dict",
            "schema": {
                "domain_name": {"required": True, "type": "string"},
                "subdomain": {"required": True, "type": "string", "allowed": ["project_id", "environment"]},
            },
        }

    return schema



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Aniek Roelofs

import pendulum
from airflow import DAG
from airflow.operators.python import (
    BranchPythonOperator,
    PythonOperator,
    ShortCircuitOperator,
)
from observatory.platform.workflows.terraform import TerraformTasks

default_args = {"owner": "airflow", "start_date": pendulum.datetime(2020, 1, 1)}

with DAG(
    dag_id=TerraformTasks.DAG_ID_DESTROY_VM,
    schedule_interval="*/10 * * * *",
    default_args=default_args,
    catchup=False,
    max_active_runs=1,
) as dag:
    # Check that dependencies exist before starting
    check = PythonOperator(
        task_id=TerraformTasks.TASK_ID_CHECK_DEPENDENCIES, python_callable=TerraformTasks.check_dependencies
    )

    # If vm is already off, stop dag
    vm_status = ShortCircuitOperator(
        task_id=TerraformTasks.TASK_ID_VM_STATUS,
        python_callable=TerraformTasks.get_variable_create,
    )

    # Destroying vm if all dags (since vm_create) are successful
    check_dags_status = BranchPythonOperator(
        task_id=TerraformTasks.TASK_ID_DAG_SUCCESS,
        python_callable=TerraformTasks.check_success_dags,
    )

    # Update terraform variable vm_create to False
    var_destroy = PythonOperator(
        task_id=TerraformTasks.TASK_ID_VAR_UPDATE,
        python_callable=TerraformTasks.update_terraform_variable,
    )

    # Run terraform configuration
    run_terraform = PythonOperator(task_id=TerraformTasks.TASK_ID_RUN, python_callable=TerraformTasks.terraform_run)

    # Check status of terraform run
    check_run_status = PythonOperator(
        task_id=TerraformTasks.TASK_ID_RUN_STATUS,
        python_callable=TerraformTasks.check_terraform_run_status,
    )

    # Check how long VM has been turned on for
    check_vm_runtime = PythonOperator(
        task_id=TerraformTasks.TASK_ID_VM_RUNTIME, python_callable=TerraformTasks.check_runtime_vm
    )

    check >> vm_status >> check_dags_status >> [var_destroy, check_vm_runtime]
    var_destroy >> run_terraform >> check_run_status



# COPYRIGHT
#
# All contributions by François Chollet:
# Copyright (c) 2015 - 2019, François Chollet.
# All rights reserved.
#
# All contributions by Google:
# Copyright (c) 2015 - 2019, Google, Inc.
# All rights reserved.
#
# All contributions by Microsoft:
# Copyright (c) 2017 - 2019, Microsoft, Inc.
# All rights reserved.
#
# All other contributions:
# Copyright (c) 2015 - 2019, the respective contributors.
# All rights reserved.
#
# Each contributor holds copyright over their respective contributions.
# The project versioning (Git) records all such contribution source information.
#
# LICENSE
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# The contents of this file originate from: https://github.com/keras-team/keras/blob/master/keras/utils/data_utils.py

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import shutil
import sys
import tarfile
import zipfile
from contextlib import closing

import six
from observatory.platform.utils.file_utils import validate_file
from six.moves.urllib.error import HTTPError
from six.moves.urllib.error import URLError
from six.moves.urllib.request import urlopen

try:
    import queue
except ImportError:
    import Queue as queue

from observatory.platform.utils.progbar_utils import Progbar

if sys.version_info[0] == 2:

    def urlretrieve(url, filename, reporthook=None, data=None):
        """Replacement for `urlretrieve` for Python 2.
        Under Python 2, `urlretrieve` relies on `FancyURLopener` from legacy
        `urllib` module, known to have issues with proxy management.
        # Arguments
            url: url to retrieve.
            filename: where to store the retrieved data locally.
            reporthook: a hook function that will be called once
                on establishment of the network connection and once
                after each block read thereafter.
                The hook will be passed three arguments;
                a count of blocks transferred so far,
                a block size in bytes, and the total size of the file.
            data: `data` argument passed to `urlopen`.
        """

        def chunk_read(response, chunk_size=8192, reporthook=None):
            content_type = response.info().get("Content-Length")
            total_size = -1
            if content_type is not None:
                total_size = int(content_type.strip())
            count = 0
            while True:
                chunk = response.read(chunk_size)
                count += 1
                if reporthook is not None:
                    reporthook(count, chunk_size, total_size)
                if chunk:
                    yield chunk
                else:
                    break

        with closing(urlopen(url, data)) as response, open(filename, "wb") as fd:
            for chunk in chunk_read(response, reporthook=reporthook):
                fd.write(chunk)


else:
    from six.moves.urllib.request import urlretrieve


def _extract_archive(file_path, path=".", archive_format="auto"):
    """Extracts an archive if it matches tar, tar.gz, tar.bz, or zip formats.
    # Arguments
        file_path: path to the archive file
        path: path to extract the archive file
        archive_format: Archive format to try for extracting the file.
            Options are 'auto', 'tar', 'zip', and None.
            'tar' includes tar, tar.gz, and tar.bz files.
            The default 'auto' is ['tar', 'zip'].
            None or an empty list will return no matches found.
    # Returns
        True if a match was found and an archive extraction was completed,
        False otherwise.
    """
    if archive_format is None:
        return False
    if archive_format == "auto":
        archive_format = ["tar", "zip"]
    if isinstance(archive_format, six.string_types):
        archive_format = [archive_format]

    for archive_type in archive_format:
        if archive_type == "tar":
            open_fn = tarfile.open
            is_match_fn = tarfile.is_tarfile
        if archive_type == "zip":
            open_fn = zipfile.ZipFile
            is_match_fn = zipfile.is_zipfile

        if is_match_fn(file_path):
            with open_fn(file_path) as archive:
                try:
                    archive.extractall(path)
                except (tarfile.TarError, RuntimeError, KeyboardInterrupt):
                    if os.path.exists(path):
                        if os.path.isfile(path):
                            os.remove(path)
                        else:
                            shutil.rmtree(path)
                    raise
            return True
    return False


def get_file(
    fname,
    origin,
    untar=False,
    md5_hash=None,
    file_hash=None,
    cache_subdir="datasets",
    hash_algorithm="auto",
    extract=False,
    archive_format="auto",
    cache_dir=None,
):
    """Downloads a file from a URL if it not already in the cache.

    By default the file at the url `origin` is downloaded to the
    cache_dir `~/.keras`, placed in the cache_subdir `datasets`,
    and given the filename `fname`. The final location of a file
    `example.txt` would therefore be `~/.keras/datasets/example.txt`.

    Files in tar, tar.gz, tar.bz, and zip formats can also be extracted.
    Passing a hash will verify the file after download. The command line
    programs `shasum` and `sha256sum` can compute the hash.

    # Arguments
        fname: Name of the file. If an absolute path `/path/to/file.txt` is
            specified the file will be saved at that location.
        origin: Original URL of the file.
        untar: Deprecated in favor of 'extract'.
            boolean, whether the file should be decompressed
        md5_hash: Deprecated in favor of 'file_hash'.
            md5 hash of the file for verification
        file_hash: The expected hash string of the file after download.
            The sha256 and md5 hash algorithms are both supported.
        cache_subdir: Subdirectory under the Keras cache dir where the file is
            saved. If an absolute path `/path/to/folder` is
            specified the file will be saved at that location.
        hash_algorithm: Select the hash algorithm to verify the file.
            options are 'md5', 'sha256', and 'auto'.
            The default 'auto' detects the hash algorithm in use.
        extract: True tries extracting the file as an Archive, like tar or zip.
        archive_format: Archive format to try for extracting the file.
            Options are 'auto', 'tar', 'zip', and None.
            'tar' includes tar, tar.gz, and tar.bz files.
            The default 'auto' is ['tar', 'zip'].
            None or an empty list will return no matches found.
        cache_dir: Location to store cached files, when None it
            defaults to the [Keras Directory](/faq/#where-is-the-keras-configuration-filed-stored).

    # Returns
        Path to the downloaded file
    """  # noqa
    if cache_dir is None:
        if "KERAS_HOME" in os.environ:
            cache_dir = os.environ.get("KERAS_HOME")
        else:
            cache_dir = os.path.join(os.path.expanduser("~"), ".keras")
    if md5_hash is not None and file_hash is None:
        file_hash = md5_hash
        hash_algorithm = "md5"
    datadir_base = os.path.expanduser(cache_dir)
    if not os.access(datadir_base, os.W_OK):
        datadir_base = os.path.join("/tmp", ".keras")
    datadir = os.path.join(datadir_base, cache_subdir)
    if not os.path.exists(datadir):
        os.makedirs(datadir)

    if untar:
        untar_fpath = os.path.join(datadir, fname)
        fpath = untar_fpath + ".tar.gz"
    else:
        fpath = os.path.join(datadir, fname)

    download = False
    if os.path.exists(fpath):
        # File found; verify integrity if a hash was provided.
        if file_hash is not None:
            if not validate_file(fpath, file_hash, algorithm=hash_algorithm):
                print(
                    "A local file was found, but it seems to be incomplete"
                    " or outdated because the {} file hash does not match "
                    "the original value of {} so we will re-download the "
                    "data.".format(hash_algorithm, file_hash)
                )
                download = True
    else:
        download = True

    if download:
        print("Downloading data from", origin)

        class ProgressTracker(object):
            # Maintain the progbar Progress Bar for the lifetime of download.
            # This design was chosen for Python 2.7 compatibility.
            progbar = None

        def dl_progress(count, block_size, total_size):
            if ProgressTracker.progbar is None:
                if total_size == -1:
                    total_size = None
                ProgressTracker.progbar = Progbar(total_size)
            else:
                ProgressTracker.progbar.update(count * block_size)

        error_msg = "URL fetch failure on {} : {} -- {}"
        try:
            try:
                urlretrieve(origin, fpath, dl_progress)
            except HTTPError as e:
                raise Exception(error_msg.format(origin, e.code, e.msg))
            except URLError as e:
                raise Exception(error_msg.format(origin, e.errno, e.reason))
        except (Exception, KeyboardInterrupt):
            if os.path.exists(fpath):
                os.remove(fpath)
            raise
        ProgressTracker.progbar = None

    if untar:
        if not os.path.exists(untar_fpath):
            _extract_archive(fpath, datadir, archive_format="tar")
        return untar_fpath

    if extract:
        _extract_archive(fpath, datadir, archive_format)

    return fpath, download



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose

import logging
import subprocess
from subprocess import Popen
from typing import List, Tuple, Union

from airflow.exceptions import AirflowException


def wait_for_process(proc: Popen) -> Tuple[str, str]:
    """Wait for a process to finish, returning the std output and std error streams as strings.

    :param proc: the process object.
    :return: std output and std error streams as strings.
    """
    output, error = proc.communicate()
    output = output.decode("utf-8")
    error = error.decode("utf-8")
    return output, error


def stream_process(proc: Popen, debug: bool) -> Tuple[str, str]:
    """Print output while a process is running, returning the std output and std error streams as strings.

    :param proc: the process object.
    :param debug: whether debug info should be displayed.
    :return: std output and std error streams as strings.
    """
    output_concat = ""
    error_concat = ""
    while True:
        for line in proc.stdout:
            output = line.decode("utf-8")
            if debug:
                print(output, end="")
            output_concat += output
        for line in proc.stderr:
            error = line.decode("utf-8")
            print(error, end="")
            error_concat += error
        if proc.poll() is not None:
            break
    return output_concat, error_concat


def run_cmd(cmd: Union[str, List[str]], shell: bool = False, executable: Union[None, str] = None):
    """Run a command (program).

    :param cmd: Command to run. Either a single string, or a list of strings.
    :param shell: Whether to use a shell to invoke it.
    :param executable: If you set shell to True, you have to specify this to the shell path, e.g., /bin/bash
    """

    p = Popen(cmd, shell=shell, stdout=subprocess.PIPE, stderr=subprocess.PIPE, executable=executable)
    stdout, stderr = wait_for_process(p)

    if stdout:
        logging.info(stdout)

    success = p.returncode == 0
    if not success:
        raise AirflowException(f"Command {cmd} failed: {stderr}")


def run_bash_cmd(cmd: str):
    """Run a command in the bash shell and wait until it's done.  Log the output.
    Raise an exception where there's stderr.

    :param cmd: Command to run.
    """

    run_cmd(cmd=cmd, shell=True, executable="/bin/bash")



# Copyright 2019 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose, Tuan Chien

import csv
import json
from typing import List


def to_json_lines(items: List[dict], serialize_custom_types_func=None) -> str:
    """Coverts a list of dictionary objects into a JSON line formatted string.

    :param items: a list of dictionaries.
    :param serialize_custom_types_func: a function that serialized types not supported by JSON by default into strings.
    :return: a JSON line formatted string.
    """

    json_records = []

    for item in items:
        json_records.append(json.dumps(item, default=serialize_custom_types_func))
        json_records.append("\n")

    json_lines = "".join(json_records)

    return json_lines


def csv_to_jsonlines(*, csv_file: str, jsonl_file: str):
    """Convert CSV file to a json lines file.

    :param csv_file: File path to csv file.
    :param jsonl_file: File path to output jsonl file.
    """

    json_lines = []

    with open(csv_file, "r") as f:
        reader = csv.DictReader(f)

        for line in reader:
            json_line = json.dumps(line) + "\n"
            json_lines.append(json_line)

    with open(jsonl_file, "w") as f:
        f.writelines(json_lines)



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose, Aniek Roelofs, Tuan Chien

import os
from typing import Tuple

import click
import stringcase
from cryptography.fernet import Fernet

from observatory.platform.observatory_config import ObservatoryConfig, TerraformConfig
from observatory.platform.utils.config_utils import module_file_path
from observatory.platform.utils.jinja2_utils import render_template


class WorkflowTypes:
    """
    Workflow types that we can generate from a template.
    """

    workflow = "Workflow"
    stream_telescope = "StreamTelescope"
    snapshot_telescope = "SnapshotTelescope"


class GenerateCommand:
    def generate_fernet_key(self) -> str:
        """Generate a Fernet key.

        :return: the Fernet key.
        """

        return Fernet.generate_key()

    def generate_local_config(self, config_path: str):
        """Command line user interface for generating an Observatory Config config.yaml.

        :param config_path: the path where the config file should be saved.
        :return: None
        """

        file_type = "Observatory Config"
        click.echo(f"Generating {file_type}...")
        ObservatoryConfig.save_default(config_path)
        click.echo(f'{file_type} saved to: "{config_path}"')

    def generate_terraform_config(self, config_path: str):
        """Command line user interface for generating a Terraform Config config-terraform.yaml.

        :param config_path: the path where the config file should be saved.
        :return: None
        """

        file_type = "Terraform Config"
        click.echo(f"Generating {file_type}...")
        TerraformConfig.save_default(config_path)
        click.echo(f'{file_type} saved to: "{config_path}"')
        click.echo(
            "Please customise the parameters with '<--' in the config file. "
            "Parameters commented out with '#' are optional."
        )

    def get_workflow_template_path_(self, workflow_type: str) -> Tuple[str, str, str, str]:
        """
        Get the correct template files to use.

        :param workflow_type: Name of the workflow type.
        :return: The workflow template path, and the dag template path.
        """

        templates_dir = module_file_path("observatory.platform.workflows.templates")
        dag_file = "workflow_dag.py.jinja2"
        test_file = "test.py.jinja2"
        doc_file = "doc.md.jinja2"

        if workflow_type == WorkflowTypes.workflow:
            workflow_file = "workflow.py.jinja2"
        elif workflow_type == WorkflowTypes.stream_telescope:
            workflow_file = "stream_telescope.py.jinja2"
        elif workflow_type == WorkflowTypes.snapshot_telescope:
            workflow_file = "snapshot_telescope.py.jinja2"
        else:
            raise Exception(f"Unsupported workflow type: {workflow_type}")

        workflow_path = os.path.join(templates_dir, workflow_file)
        dag_path = os.path.join(templates_dir, dag_file)
        test_path = os.path.join(templates_dir, test_file)
        doc_path = os.path.join(templates_dir, doc_file)
        return workflow_path, dag_path, test_path, doc_path

    def generate_new_workflow(self, *, project_path: str, package_name: str, workflow_type: str, workflow_name: str):
        """Make a new workflow template.

        :param project_path: the path to the workflows project.
        :param package_name: the Python package name.
        :param workflow_type: Type of workflow to generate.
        :param workflow_name: Class name of the new telescope.
        """

        # Standardise names
        package_name = stringcase.snakecase(package_name)
        workflow_class = stringcase.capitalcase(stringcase.camelcase(workflow_name))
        workflow_module = stringcase.lowercase(stringcase.snakecase(workflow_name))
        workflow_file = f"{workflow_module}.py"

        # Destination folders e.g.
        # academic-observatory-workflows/academic_observatory_workflows/workflows
        # academic-observatory-workflows/academic_observatory_workflows/dags
        # academic-observatory-workflows/tests/academic_observatory_workflows/workflows
        # academic-observatory-workflows/docs
        workflow_dst_dir = os.path.join(project_path, package_name, "workflows")
        dags_dst_dir = os.path.join(project_path, package_name, "dags")
        test_dst_dir = os.path.join(project_path, "tests", package_name, "workflows")
        doc_dst_dir = os.path.join(project_path, "docs")

        # Make folders
        for path in [workflow_dst_dir, dags_dst_dir, test_dst_dir, doc_dst_dir]:
            os.makedirs(path, exist_ok=True)

        # Make init files
        package_folder = os.path.join(project_path, package_name)
        tests_package_folder = os.path.join(project_path, "tests", package_name)
        init_paths = [package_folder, tests_package_folder, workflow_dst_dir, dags_dst_dir, test_dst_dir]
        for path in init_paths:
            if not os.path.isfile(path):
                open(os.path.join(path, "__init__.py"), "a").close()

        # Destination files
        workflow_dst_file = os.path.join(workflow_dst_dir, workflow_file)
        dag_dst_file = os.path.join(dags_dst_dir, workflow_file)
        test_dst_file = os.path.join(test_dst_dir, f"test_{workflow_file}")
        doc_dst_file = os.path.join(doc_dst_dir, f"{workflow_module}.md")
        doc_index_file = os.path.join(doc_dst_dir, f"index.rst")

        # Get template paths
        (
            workflow_template_path,
            dag_template_path,
            test_template_path,
            doc_template_path,
        ) = self.get_workflow_template_path_(workflow_type)

        # Render files
        workflow = render_template(
            workflow_template_path, workflow_class=workflow_class, workflow_module=workflow_module
        )
        dag = render_template(
            dag_template_path,
            package_name=package_name,
            workflow_class=workflow_class,
            workflow_module=workflow_module,
        )
        test = render_template(test_template_path, workflow_class=workflow_class, workflow_module=workflow_module)
        doc = render_template(doc_template_path, workflow_class=workflow_class, workflow_module=workflow_module)

        # Write out dag template
        with open(dag_dst_file, "w") as f:
            f.write(dag)
        print(f"Created a new workflow file: {workflow_dst_file}")

        # Write out workflow template
        with open(workflow_dst_file, "w") as f:
            f.write(workflow)
        print(f"Created a new dag file: {dag_dst_file}")

        # Write out test template
        with open(test_dst_file, "w") as f:
            f.write(test)
        print(f"Created a new workflow documentation file: {doc_dst_file}")

        # Write out documentation template
        with open(doc_dst_file, "w") as f:
            f.write(doc)
        print(f"Created a new workflow test file: {test_dst_file}")

        # Add the new workflow doc to the documentation index
        with open(doc_index_file, "a") as f:
            f.write(f"    {workflow_module}\n")
        print(f"Add the new workflow doc to the documentation index: {test_dst_file}")



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose, Aniek Roelofs, Tuan Chien

import os

import click

from observatory.platform.cli.click_utils import INDENT1, INDENT2, INDENT3, indent
from observatory.platform.cli.generate_command import GenerateCommand
from observatory.platform.cli.platform_command import PlatformCommand
from observatory.platform.cli.terraform_command import TerraformCommand
from observatory.platform.platform_builder import DEBUG, HOST_GID, HOST_UID
from observatory.platform.utils.config_utils import (
    terraform_credentials_path as default_terraform_credentials_path,
    observatory_home,
)

PLATFORM_NAME = "Observatory Platform"
TERRAFORM_NAME = "Observatory Terraform"

LOCAL_CONFIG_PATH = os.path.join(observatory_home(), "config.yaml")
TERRAFORM_CONFIG_PATH = os.path.join(observatory_home(), "config-terraform.yaml")


@click.group()
def cli():
    """The Observatory Platform command line tool.

    COMMAND: the commands to run include:\n
      - platform: start and stop the local Observatory Platform platform.\n
      - generate: generate a variety of outputs.\n
      - terraform: manage Terraform Cloud workspaces.\n
    """

    pass


@cli.command()
@click.argument("command", type=click.Choice(["start", "stop"]))
@click.option(
    "--config-path",
    type=click.Path(exists=False, file_okay=True, dir_okay=False),
    default=LOCAL_CONFIG_PATH,
    help="The path to the config.yaml configuration file.",
    show_default=True,
)
@click.option(
    "--host-uid",
    type=click.INT,
    default=HOST_UID,
    help="The user id of the host system. Used to set the user id in the Docker containers.",
    show_default=True,
)
@click.option(
    "--host-gid",
    type=click.INT,
    default=HOST_GID,
    help="The group id of the host system. Used to set the group id in the Docker containers.",
    show_default=True,
)
@click.option("--debug", is_flag=True, default=DEBUG, help="Print debugging information.")
def platform(
    command: str,
    config_path: str,
    host_uid: int,
    host_gid: int,
    debug,
):
    """Run the local Observatory Platform platform.\n

    COMMAND: the command to give the platform:\n
      - start: start the platform.\n
      - stop: stop the platform.\n
    """

    min_line_chars = 80
    print(f"{PLATFORM_NAME}: checking dependencies...".ljust(min_line_chars), end="\r")
    if os.path.isfile(config_path):
        # Make the platform command, which encapsulates functionality for running the observatory
        platform_cmd = PlatformCommand(
            config_path,
            host_uid=host_uid,
            host_gid=host_gid,
            debug=debug,
        )

        # Check dependencies
        platform_check_dependencies(platform_cmd, min_line_chars=min_line_chars)

        # Start the appropriate process
        if command == "start":
            platform_start(platform_cmd)
        elif command == "stop":
            platform_stop(platform_cmd)

        exit(os.EX_OK)
    else:
        print(indent("config.yaml:", INDENT1))
        print(indent(f"- file not found, generating a default file on path: {config_path}", INDENT2))
        generate_cmd = GenerateCommand()
        generate_cmd.generate_local_config(config_path)
        exit(os.EX_CONFIG)


def platform_check_dependencies(platform_cmd: PlatformCommand, min_line_chars: int = 80):
    """Check Platform dependencies.

    :param platform_cmd: the platform command instance.
    :param min_line_chars: the minimum number of lines when printing to the command line interface.
    :return: None.
    """

    if not platform_cmd.is_environment_valid:
        print(f"{PLATFORM_NAME}: dependencies missing".ljust(min_line_chars))
    else:
        print(f"{PLATFORM_NAME}: all dependencies found".ljust(min_line_chars))

    print(indent("Docker:", INDENT1))
    if platform_cmd.docker_exe_path is not None:
        print(indent(f"- path: {platform_cmd.docker_exe_path}", INDENT2))

        if platform_cmd.is_docker_running:
            print(indent(f"- running", INDENT2))
        else:
            print(indent("- not running, please start", INDENT2))
    else:
        print(indent("- not installed, please install https://docs.docker.com/get-docker/", INDENT2))

    print(indent("Docker Compose:", INDENT1))
    if platform_cmd.docker_compose_path is not None:
        print(indent(f"- path: {platform_cmd.docker_compose_path}", INDENT2))
    else:
        print(indent("- not installed, please install https://docs.docker.com/compose/install/", INDENT2))

    print(indent("config.yaml:", INDENT1))
    print(indent(f"- path: {platform_cmd.config_path}", INDENT2))
    if platform_cmd.config.is_valid:
        print(indent("- file valid", INDENT2))
    else:
        print(indent("- file invalid", INDENT2))
        for error in platform_cmd.config.errors:
            print(indent("- {}: {}".format(error.key, error.value), INDENT3))

    if not platform_cmd.is_environment_valid:
        exit(os.EX_CONFIG)

    print(indent("Host machine settings:", INDENT1))
    print(indent(f"- observatory home: {platform_cmd.config.observatory.observatory_home}", INDENT2))
    print(indent(f"- host-uid: {platform_cmd.host_uid}", INDENT2))


def platform_start(platform_cmd: PlatformCommand, min_line_chars: int = 80):
    """Check Platform dependencies.

    :param platform_cmd: the platform command instance.
    :param min_line_chars: the minimum number of lines when printing to the command line interface.
    :return: None.
    """

    print(f"{PLATFORM_NAME}: building...".ljust(min_line_chars), end="\r")
    response = platform_cmd.build()

    if response.return_code == 0:
        print(f"{PLATFORM_NAME}: built".ljust(min_line_chars))
    else:
        print(f"{PLATFORM_NAME}: build error".ljust(min_line_chars))
        exit(os.EX_CONFIG)

    # Start the built containers
    print(f"{PLATFORM_NAME}: starting...".ljust(min_line_chars), end="\r")
    response = platform_cmd.start()

    if response.return_code == 0:
        ui_started = platform_cmd.wait_for_airflow_ui(timeout=120)

        if ui_started:
            print(f"{PLATFORM_NAME}: started".ljust(min_line_chars))
            print(f"View the Apache Airflow UI at {platform_cmd.ui_url}")
        else:
            print(f"{PLATFORM_NAME}: error starting".ljust(min_line_chars))
            print(f"Could not find the Airflow UI at {platform_cmd.ui_url}")
    else:
        print("Error starting the Observatory Platform")
        exit(os.EX_CONFIG)


def platform_stop(platform_cmd: PlatformCommand, min_line_chars: int = 80):
    """Start the Observatory platform.

    :param platform_cmd: the platform command instance.
    :param min_line_chars: the minimum number of lines when printing to the command line interface.
    :return: None.
    """

    print(f"{PLATFORM_NAME}: stopping...".ljust(min_line_chars), end="\r")
    response = platform_cmd.stop()

    if platform_cmd.debug:
        print(response.output)

    if response.return_code == 0:
        print(f"{PLATFORM_NAME}: stopped".ljust(min_line_chars))
    else:
        print(f"{PLATFORM_NAME}: error stopping".ljust(min_line_chars))
        print(response.error)
        exit(os.EX_CONFIG)


@cli.group()
def generate():
    """The Observatory Platform generate command.

    COMMAND: the commands to run include:\n
      - secrets: generate secrets.\n
      - config: generate configuration files for the Observatory Platform.\n
    """

    pass


@generate.command()
@click.argument(
    "project_path",
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
)
@click.argument("package_name", type=str)
@click.argument("workflow_type", type=click.Choice(["Workflow", "StreamTelescope", "SnapshotTelescope"]))
@click.argument("workflow_name", type=str)
def workflow(project_path: str, package_name: str, workflow_type: str, workflow_name: str):
    """Generate a new workflow.

    \b
    PROJECT_PATH is the Python project path.
    PACKAGE_NAME is the Python package name.
    WORKFLOW_TYPE is the type of workflow.
    WORKFLOW_NAME is the the name of your new workflow.

    \b
    For example, the command: observatory generate workflow /path/to/my-workflows-project my_workflows_project Workflow MyWorkflow

    \b Will generate the following files and folders:

    \b
    └── my-workflows-project
        ├── docs
        │   ├── index.rst
        │   └── my_workflow.md
        ├── my_workflows_project
        │   ├── dags
        │   │   ├── __init__.py
        │   │   └── my_workflow.py
        │   ├── __init__.py
        │   └── workflows
        │       ├── __init__.py
        │       └── my_workflow.py
        └── tests
            └── my_workflows_project
                ├── __init__.py
                └── workflows
                    ├── __init__.py
                    └── test_my_workflow.py
    """

    # TODO: project path and package name should be inferred somehow
    cmd = GenerateCommand()
    cmd.generate_new_workflow(
        project_path=project_path, package_name=package_name, workflow_type=workflow_type, workflow_name=workflow_name
    )


@generate.command()
@click.argument("command", type=click.Choice(["fernet-key"]))
def secrets(command: str):
    """Generate secrets for the Observatory Platform.\n

    COMMAND: the type of secret to generate:\n
      - fernet-key: generate a random Fernet Key.\n
    """

    if command == "fernet-key":
        cmd = GenerateCommand()
        print(cmd.generate_fernet_key())


@generate.command()
@click.argument("command", type=click.Choice(["local", "terraform"]))
@click.option(
    "--config-path",
    type=click.Path(exists=False, file_okay=True, dir_okay=False),
    default=None,
    help="The path to the config file to generate.",
    show_default=True,
)
def config(command: str, config_path: str):
    """Generate config files for the Observatory Platform.\n

    COMMAND: the type of config file to generate:\n
      - local: generate a config file for running the Observatory Platform locally.\n
      - terraform: generate a config file for running the Observatory Platform with Terraform.\n
    """

    # Make the generate command, which encapsulates functionality for generating data
    cmd = GenerateCommand()

    cmd_func = None
    config_name = ""
    if command == "local":
        if config_path is None:
            config_path = LOCAL_CONFIG_PATH
        cmd_func = cmd.generate_local_config
        config_name = "Observatory Config"
    elif command == "terraform":
        if config_path is None:
            config_path = TERRAFORM_CONFIG_PATH
        cmd_func = cmd.generate_terraform_config
        config_name = "Terraform Config"

    if not os.path.exists(config_path) or click.confirm(
        f'The file "{config_path}" exists, do you want to overwrite it?'
    ):
        cmd_func(config_path)
    else:
        click.echo(f"Not generating {config_name}")


# increase content width for cleaner help output
@cli.command(context_settings=dict(max_content_width=120))
@click.argument(
    "command",
    type=click.Choice(["build-terraform", "build-image", "build-api-image", "create-workspace", "update-workspace"]),
)
# The path to the config-terraform.yaml configuration file.
@click.argument("config-path", type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.option(
    "--terraform-credentials-path",
    type=click.Path(exists=False, file_okay=True, dir_okay=False),
    default=default_terraform_credentials_path(),
    help="",
    show_default=True,
)
@click.option("--debug", is_flag=True, default=DEBUG, help="Print debugging information.")
def terraform(command, config_path, terraform_credentials_path, debug):
    """Commands to manage the deployment of the Observatory Platform with Terraform Cloud.\n

    COMMAND: the type of config file to generate:\n
      - create-workspace: create a Terraform Cloud workspace.\n
      - update-workspace: update a Terraform Cloud workspace.\n
      - build-image: build a Google Compute image for the Terraform deployment with Packer.\n
      - build-terraform: build the Terraform files.\n
    """

    # The minimum number of characters per line
    min_line_chars = 80

    terraform_cmd = TerraformCommand(config_path, terraform_credentials_path, debug=debug)
    generate_cmd = GenerateCommand()

    # Check dependencies
    terraform_check_dependencies(terraform_cmd, generate_cmd)

    # Run commands
    if command == "build-terraform":
        # Build image with packer
        terraform_cmd.build_terraform()
    elif command == "build-image":
        # Build image with packer
        terraform_cmd.build_image()
    elif command == "build-api-image":
        # Build docker image stored in google container registry
        terraform_cmd.build_google_container_image()
    else:
        # Create a new workspace
        if command == "create-workspace":
            terraform_cmd.create_workspace()

        # Update an existing workspace
        elif command == "update-workspace":
            terraform_cmd.update_workspace()


def terraform_check_dependencies(
    terraform_cmd: TerraformCommand, generate_cmd: GenerateCommand, min_line_chars: int = 80
):
    """Check Terraform dependencies.

    :param terraform_cmd: the Terraform command instance.
    :param generate_cmd: the generate command instance.
    :param min_line_chars: the minimum number of lines when printing to the command line interface.
    :return: None.
    """

    print(f"{TERRAFORM_NAME}: checking dependencies...".ljust(min_line_chars), end="\r")

    if not terraform_cmd.is_environment_valid:
        print(f"{TERRAFORM_NAME}: dependencies missing".ljust(min_line_chars))
    else:
        print(f"{TERRAFORM_NAME}: all dependencies found".ljust(min_line_chars))

    print(indent("Config:", INDENT1))
    if terraform_cmd.config_exists:
        print(indent(f"- path: {terraform_cmd.config_path}", INDENT2))
        if terraform_cmd.config.is_valid:
            print(indent("- file valid", INDENT2))
        else:
            print(indent("- file invalid", INDENT2))
            for key, value in terraform_cmd.config.validator.errors.items():
                print(indent(f"- {key}: {value}", INDENT3))
    else:
        print(indent("- file not found, generating a default file", INDENT2))
        generate_cmd.generate_terraform_config(terraform_cmd.config_path)

    print(indent("Terraform credentials file:", INDENT1))
    if terraform_cmd.terraform_credentials_exists:
        print(indent(f"- path: {terraform_cmd.terraform_credentials_path}", INDENT2))
    else:
        print(indent("- file not found, create one by running 'terraform login'", INDENT2))

    print(indent("Packer", INDENT1))
    if terraform_cmd.terraform_builder.packer_exe_path is not None:
        print(indent(f"- path: {terraform_cmd.terraform_builder.packer_exe_path}", INDENT2))
    else:
        print(indent("- not installed, please install https://www.packer.io/docs/install", INDENT2))

    print(indent("Google Cloud SDK", INDENT1))
    if terraform_cmd.terraform_builder.gcloud_exe_path is not None:
        print(indent(f"- path: {terraform_cmd.terraform_builder.gcloud_exe_path}", INDENT2))
    else:
        print(indent("- not installed, please install https://cloud.google.com/sdk/docs/install", INDENT2))

    if not terraform_cmd.is_environment_valid:
        exit(os.EX_CONFIG)


if __name__ == "__main__":
    cli()
